"""
textDocument/documentSymbol
"""
#  Copyright (c) 2019 Seven Bridges. See LICENSE

import pathlib
import json
import hashlib

from .base import CWLLangServerBase

import logging
logger = logging.getLogger(__name__)


class DocumentSymbol(CWLLangServerBase):

    def serve_textDocument_documentSymbol(self, client_query):
        params = client_query["params"]
        doc_uri = params["textDocument"]["uri"]

        doc = self.open_documents[doc_uri]

        self._write_out_graph(doc)
        return doc.symbols

    def _write_out_graph(self, doc):
        graph_data_file = pathlib.Path(
            self.config.scratch_path,
            hashlib.md5(doc.doc_uri.encode()).hexdigest() + ".json")
        with graph_data_file.open("w") as f:
            json.dump(doc.wf_graph, f, indent=2)



#  Copyright (c) 2019 Seven Bridges. See LICENSE

from .intelligence import IntelligenceNode
from .intelligencecontext import IntelligenceContext


class Requirements(IntelligenceContext):

    def __init__(self, req_types):
        self.req_types = req_types

    def get_completer(self):
        return IntelligenceNode(completions=self.req_types)



#  Copyright (c) 2019 Seven Bridges. See LICENSE

from .basetype import CWLBaseType, MapSubjectPredicate, TypeCheck

import logging
logger = logging.getLogger(__name__)


class CWLAnyType(CWLBaseType):

    def __init__(self, name, type_dict):
        super().__init__(name)
        self.type_dict = type_dict
        # The Any type is special because it can be asked to
        # be anything. We need to carry the entire type dict
        # with us.

    def if_you_can_be_anything_be_this_kind(self, type_name):
        return self.type_dict.get(type_name)

    def all_possible_type_names(self):
        return list(self.type_dict.keys())

    def check(self, node, node_key: str=None, map_sp: MapSubjectPredicate=None) -> TypeCheck:
        # Special treatment for the any type. It agrees to everything
        return TypeCheck(self)



#  Copyright (c) 2019 Seven Bridges. See LICENSE

from .basetype import CWLBaseType, Intelligence, MapSubjectPredicate, TypeCheck, Match
from ..langserver.lspobjects import Range
from ..code.intelligencecontext import IntelligenceContext
from .typeinference import infer_type
from .lib import get_range_for_value


class CWLArrayType(CWLBaseType):

    def __init__(self, name, allowed_types):
        super().__init__(name)
        if not isinstance(allowed_types, list):
            allowed_types = [allowed_types]
        self.types = allowed_types

    def check(self, node, node_key: str=None, map_sp: MapSubjectPredicate=None) -> TypeCheck:
        if isinstance(node, list):
            return TypeCheck(self)
        else:
            return TypeCheck(self, match=Match.No)

    def parse(self,
              doc_uri: str,
              node,
              intel_context: IntelligenceContext,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):

        for n, v in enumerate(node):
            inferred_type = infer_type(v, self.types)
            inferred_type.parse(
                doc_uri=doc_uri,
                node=v,
                intel_context=intel_context,
                code_intel=code_intel,
                problems=problems,
                value_range=get_range_for_value(node, n),
                requirements=requirements)



"""Code to load the CWL specification from JSON and represent it as a
set of types"""

#  Copyright (c) 2019 Seven Bridges. See LICENSE

import json

from .alltypes import *


import logging
logger = logging.getLogger(__name__)


latest_published_cwl_version = "v1.1"
process_types = ["CommandLineTool", "ExpressionTool", "Workflow"]


def parse_schema(fname):
    type_dict = {}
    schema = json.load(open(fname, "r"))

    # we need multiple passes to resolve forward references
    # todo: a way to determine how many passes are needed
    for _ in range(3):
        parse_cwl_type(schema, type_dict)
        add_formal_primitive_types_to_type_dict(type_dict)

    clean_up_schema(type_dict)
    return type_dict


def add_formal_primitive_types_to_type_dict(type_dict):
    for pt in type_dict.get("PrimitiveType").symbols:
        type_dict[pt] = CWLBaseType(name=pt)


# CWL grows hairy. It needs to be shaved periodically
def clean_up_schema(type_dict):
    if isinstance(type_dict.get("CWLVersion"), CWLEnumType):
        type_dict["CWLVersion"].symbols = [
            symbol
            for symbol in type_dict["CWLVersion"].symbols
            if not symbol.startswith("draft-")
        ]
        type_dict["CWLVersion"].symbols.remove("v1.0.dev4")
        return

    logger.error("No CWLVersion enum in schema")


def parse_cwl_type(schema, lang_model, map_subject_predicate=None, field_name=None):

    if isinstance(schema, str):
        return lang_model.get(schema, schema)

    elif isinstance(schema, list):
        return [
            parse_cwl_type(_scheme, lang_model, map_subject_predicate, field_name)
            for _scheme in schema
        ]

    elif schema.get("type") == "array":
        name = field_name  # schema.get("name")
        if map_subject_predicate is None:
            return CWLArrayType(
                name=name,
                allowed_types=parse_cwl_type(schema.get("items"), lang_model))
        else:
            return CWLListOrMapType(
                name=name,
                allowed_types=parse_cwl_type(schema.get("items"), lang_model),
                map_sp=map_subject_predicate)

    elif schema.get("type") == "enum":
        return parse_enum(schema, lang_model)

    elif schema.get("type") == "record":
        return parse_record(schema, lang_model)

    logger.error(f"Unknown schema type {schema.get('type')}: {schema}")


def parse_enum(schema, lang_model):
    enum_name = schema.get("name")

    if enum_name == "Any":
        # This is a special type masquerading as an enum
        lang_model[enum_name] = CWLAnyType(name=schema.get("name"), type_dict=lang_model)
        return lang_model.get(enum_name)

    if enum_name == "Expression":
        # Yet another special type masquerading as an enum
        lang_model[enum_name] = CWLExpressionType(name=schema.get("name"))
        return lang_model.get(enum_name)

    symbols = schema.get("symbols")
    for extends in listify(schema.get("extends")):
        extends = extends.split("#")[1]  # Proper way is to use URIs ...
        if extends in lang_model:
            symbols += lang_model[extends].symbols

    lang_model[enum_name] = CWLEnumType(
        name=schema.get("name"),
        symbols=set(symbols))

    return lang_model.get(enum_name)


def parse_record(schema, lang_model):
    record_name = schema.get("name")
    fields = {}
    for extends in listify(schema.get("extends")):
        extends = extends.split("#")[1]  # Proper way is to use URIs ...
        if extends in lang_model:
            fields.update(**lang_model[extends].fields)

    fields.update(**{
        k: v
        for field in schema.get("fields") for k, v in [parse_field(field, lang_model)]
    })

    lang_model[record_name] = CWLRecordType(
        name=record_name,
        doc=schema.get("doc"),
        fields=fields)

    return lang_model.get(record_name)


def listify(items):
    if items is None:
        return []
    elif isinstance(items, list):
        return items
    else:
        return [items]


def parse_field(field, lang_model):

    field_name = field.get("name")
    required = True
    _allowed_types = field.get("type")
    if isinstance(_allowed_types, list) and "null" in _allowed_types:
        required = False

    map_subject_predicate = None
    jldp = field.get("jsonldPredicate")
    if isinstance(jldp, dict):
        if "mapSubject" in jldp:
            map_subject_predicate = MapSubjectPredicate(jldp.get("mapSubject"),
                                                        jldp.get("mapPredicate"))

    return field_name, CWLFieldType(
        doc=field.get("doc"),
        required=required,
        allowed_types=parse_cwl_type(
            _allowed_types, lang_model,
            map_subject_predicate=map_subject_predicate,
            field_name=field_name)
    )



#  Copyright (c) 2019 Seven Bridges. See LICENSE

from .basetype import CWLBaseType, IntelligenceContext, Intelligence, MapSubjectPredicate, TypeCheck, Match
from .unknowntype import CWLUnknownType
from .requirementstype import CWLRequirementsType
from ..langserver.lspobjects import Range
from ..code.intelligence import LookupNode, IntelligenceNode
from ..code.intelligencecontext import copy_context
from .lib import ListOrMap
from .typeinference import infer_type
from ..code import workflow


# TODO: subclass this to handle
#  inputs and outputs re: filling out keys = ports
#  requirements re: filling out types
#  To do this we need to properly propagate the name of the field to LOM types
#  and to properly handle the broken document that results when we start to fill
#  out a key

# A list or map type is almost always interesting (see docs/document-model.md)
# Requirements, step `in` and workflow `output` fields have special completions
# So for each of these we instantiate specialized types for the completers
# Parsing proceeds normally

class CWLListOrMapType(CWLBaseType):

    def __init__(self, name, allowed_types, map_sp: MapSubjectPredicate):
        super().__init__(name)
        self.map_subject_predicate = map_sp
        if not isinstance(allowed_types, list):
            allowed_types = [allowed_types]
        self.types = allowed_types
        self.enclosing_workflow = None

    def check(self, node, node_key: str=None, map_sp: MapSubjectPredicate=None) -> TypeCheck:
        if node is None or isinstance(node, (str, list, dict)):
            return TypeCheck(cwl_type=self)
        else:
            return TypeCheck(cwl_type=self, match=Match.No)

    def parse(self,
              doc_uri: str,
              node,
              intel_context: IntelligenceContext,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):

        obj = ListOrMap(node, key_field=self.map_subject_predicate.subject, problems=problems)

        # items expressed as a map can have a special case
        # the key completer can be the completer for the subject field
        # if the value is a string, it's completer is the completer for the predicate field
        if self.name == "requirements":
            intel_context.requirements = IntelligenceNode(completions=[t.name for t in self.types])

        # Exception, when we have an empty LOM we assume it's gonna be a dict and
        # allow completions.
        # We end up with None or str when we just start a lom
        if obj.original_obj is None or isinstance(obj.original_obj, str):
            # TODO: remove redundancy
            # The keys get fancy completions
            if self.name == "requirements":
                ln = LookupNode(loc=value_range)
                ln.intelligence_node = intel_context.requirements
                code_intel.add_lookup_node(ln)

            elif self.name == "in":
                if intel_context.workflow_step_intelligence is not None:
                    ln = LookupNode(loc=value_range)
                    ln.intelligence_node = intel_context.workflow_step_intelligence.get_step_inport_completer()
                    code_intel.add_lookup_node(ln)

            elif self.name == "output":
                if intel_context.workflow is not None:
                    ln = LookupNode(loc=value_range)
                    ln.intelligence_node = intel_context.workflow.get_output_source_completer("")
                    code_intel.add_lookup_node(ln)

        for k, v in obj.as_dict.items():

            this_intel_context = copy_context(intel_context)
            this_intel_context.path += [k]

            inferred_type = infer_type(
                v,
                allowed_types=self.types,
                key=k if obj.was_dict else None,
                map_sp=self.map_subject_predicate if obj.was_dict else None)

            if self.name == "requirements" and isinstance(inferred_type, CWLUnknownType):
                # We've begun to type, and we can't formally infer this is a Requirement
                # but we'd like to offer auto-completions of requirement types
                inferred_type = CWLRequirementsType("requirement", self.types)

            if self.name == "steps":
                this_intel_context.workflow_step_intelligence = workflow.WFStepIntelligence(step_id=k)

            inferred_type.parse(
                doc_uri=doc_uri,
                node=v,
                intel_context=this_intel_context,
                code_intel=code_intel,
                problems=problems,
                node_key=k if obj.was_dict else None,
                map_sp=self.map_subject_predicate,
                key_range=obj.get_range_for_id(k),
                value_range=obj.get_range_for_value(k),
                requirements=requirements)

            if self.name == "steps":
                intel_context.workflow.add_step_intel(k, this_intel_context.workflow_step_intelligence)

            if obj.was_dict:
                # The keys get fancy completions
                if self.name == "requirements":
                    ln = LookupNode(loc=obj.get_range_for_id(k))
                    ln.intelligence_node = intel_context.requirements
                    code_intel.add_lookup_node(ln)

                elif self.name == "in":

                    wf_step = intel_context.workflow_step_intelligence
                    if wf_step is not None:

                        ln = LookupNode(loc=obj.get_range_for_id(k))
                        ln.intelligence_node = wf_step.get_step_inport_completer()
                        code_intel.add_lookup_node(ln)

                        if v is None or isinstance(v, str):
                            ln = LookupNode(loc=obj.get_range_for_value(k))
                            ln.intelligence_node = wf_step.get_step_source_completer(v)
                            code_intel.add_lookup_node(ln)

                elif self.name == "output":
                    if v is None or isinstance(v, str):
                        ln = LookupNode(loc=obj.get_range_for_value(k))
                        ln.intelligence_node = intel_context.workflow.get_output_source_completer(v)
                        code_intel.add_lookup_node(ln)



"""
textDocument/hover
"""
#  Copyright (c) 2019 Seven Bridges. See LICENSE

from .lspobjects import Position
from .base import CWLLangServerBase

import logging
logger = logging.getLogger(__name__)


class Hover(CWLLangServerBase):

    def serve_textDocument_hover(self, client_query):
        params = client_query["params"]
        doc_uri = params["textDocument"]["uri"]
        position = Position(**params["position"])
        doc = self.open_documents[doc_uri]

        return doc.hover(position)



#  Copyright (c) 2019 Seven Bridges. See LICENSE

from typing import Dict

from .basetype import CWLBaseType, IntelligenceContext, Intelligence, MapSubjectPredicate, TypeCheck, Match
from .linkedfiletype import CWLLinkedFile
from .linkedschemadeftype import CWLLinkedSchemaDef
from .namespacedtype import CWLNameSpacedType
from .expressiontype import CWLExpression
from ..langserver.lspobjects import Range, CompletionItem, Diagnostic, DiagnosticSeverity
from ..code.intelligence import LookupNode
from ..code.intelligencecontext import copy_context
from ..code.workflow import Workflow
from .typeinference import infer_type
from .lib import get_range_for_key, get_range_for_value
from ..code import workflow

import logging
logger = logging.getLogger(__name__)


class CWLRecordType(CWLBaseType):

    def __init__(self, name: str, doc: str, fields: Dict[str, 'CWLFieldType']):
        super().__init__(name)
        self.doc = doc
        self.fields = fields
        self.required_fields = set((k for k, v in self.fields.items() if v.required))
        self.all_fields = set(self.fields.keys())

    def check(self, node, node_key: str=None, map_sp: MapSubjectPredicate=None) -> TypeCheck:

        if node is None:
            return TypeCheck(self, Match.No)

        # Exception for $import/$include etc.
        if isinstance(node, dict):
            if "$import" in node:
                if self.name == "InputRecordSchema":
                    return TypeCheck(cwl_type=CWLLinkedSchemaDef(node.get("$import")))
                else:
                    return TypeCheck(cwl_type=CWLLinkedFile(node.get("$import")))

        required_fields = self.required_fields - {map_sp.subject if map_sp else None}

        if not isinstance(node, dict):
            if map_sp is not None and node_key is not None:
                if map_sp.predicate in self.required_fields and len(required_fields) <= 1:
                    return TypeCheck(cwl_type=self)

            return TypeCheck(cwl_type=self, match=Match.No)

        fields_present = set(node.keys())
        missing_fields = required_fields - fields_present
        extra_fields = fields_present - self.all_fields
        if len(missing_fields):
            return TypeCheck(cwl_type=self,
                             match=Match.Maybe, missing_req_fields=list(missing_fields))
        elif len(extra_fields):
            return TypeCheck(cwl_type=self,
                             match=Match.Maybe)
        else:
            return TypeCheck(cwl_type=self)

    def parse(self,
              doc_uri: str,
              node,
              intel_context: IntelligenceContext,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):

        if not isinstance(node, dict):
            if map_sp is not None and map_sp.predicate is not None:
                _field_iterator = [(map_sp.predicate, node)]
            else:
                return
        else:
            _field_iterator = node.items()

        field_iterator = _put_this_field_first(_field_iterator, "requirements")
        # We need to process the requirements first so that we resolve typedefs and JS libraries

        field_iterator = _put_this_field_first(field_iterator, "when")
        # The when field may use input ports not present in the underlying tool
        # we process this first and have the extra inputs at hand so we can add them
        # to the step interface
        extra_inputs_for_when = []

        if self.name == "Workflow":
            intel_context.workflow = Workflow(node.get("inputs"), node.get("outputs"), node.get("steps"))

        for k, child_node in field_iterator:

            inferred_type = None
            this_intel_context = copy_context(intel_context)
            this_intel_context.path += [k]

            if isinstance(node, dict):
                key_range = get_range_for_key(node, k)
                value_range = get_range_for_value(node, k)

                # key completer
                ln = LookupNode(loc=key_range)
                ln.intelligence_node = self
                code_intel.add_lookup_node(ln)

            # TODO: looks like this logic and the logic in lomtype can be combined
            # Special completers

            # These paths tend to look like
            # ['requirements', 'XXRequirement', 'class']
            if k == "class" and len(this_intel_context.path) > 2 and \
                    this_intel_context.path[-3] == "requirements":
                ln = LookupNode(loc=value_range)
                ln.intelligence_node = this_intel_context.requirements
                code_intel.add_lookup_node(ln)

            if self.name == "WorkflowStep" and k == "run" and isinstance(child_node, str):
                # Exception for run field that is a string
                inferred_type = CWLLinkedFile(prefix=child_node, extension=".cwl")

            elif self.name == "InlineJavascriptRequirement" and k == "expressionLib":
                # todo: this will fail for inlined nested workflows
                code_intel.prepare_expression_lib(child_node)
                continue

            elif k in ["$schemas", "$namespaces"]:
                continue

            elif ":" in k:
                inferred_type = CWLNameSpacedType(k)

            else:

                # Regular processing
                field = self.fields.get(k)
                if field is None:
                    if ":" not in k and k[0] != "$":
                        # heuristics to ignore $schemas, $namespaces and custom tags
                        problems += [
                            Diagnostic(
                                _range=key_range,
                                message=f"Unknown field: {k} for type {self.name}",
                                severity=DiagnosticSeverity.Warning)
                        ]
                    continue

                else:
                    inferred_type = infer_type(child_node, field.types, key=k)

            inferred_type.parse(
                doc_uri=doc_uri,
                node=child_node,
                intel_context=this_intel_context,
                code_intel=code_intel,
                problems=problems,
                node_key=k,
                map_sp=map_sp,
                key_range=key_range,
                value_range=value_range,
                requirements=requirements)

            if self.name == "WorkflowStep" and k == "when" \
                    and isinstance(inferred_type, CWLExpression):
                extra_inputs_for_when = inferred_type.guess_inputs()

            if self.name == "WorkflowOutputParameter" and k == "outputSource":
                set_port_completers(code_intel, child_node, value_range,
                                    this_intel_context.workflow.get_output_source_completer)

            if self.name == "WorkflowStepInput" and k == "source":
                set_port_completers(code_intel, child_node, value_range,
                                    this_intel_context.workflow_step_intelligence.get_step_source_completer)

            if self.name == "WorkflowStep" and k == "run":
                if isinstance(inferred_type, CWLLinkedFile):
                    linked_process = inferred_type.node_dict
                else:
                    linked_process = child_node

                step_interface = workflow.parse_step_interface(linked_process, problems)
                step_interface.inputs.update(extra_inputs_for_when)
                intel_context.workflow_step_intelligence.set_step_interface(step_interface)

        if self.name == "Workflow":
            intel_context.workflow.validate_connections(problems=problems)

    def completion(self):
        return [CompletionItem(label=k) for k in self.fields.keys()]


# Our sources can be singular or a list: we need to handle both
def set_port_completers(code_intel, node, value_range, get_source_completer):
    if isinstance(node, list):
        _iterator = ((_node, get_range_for_value(node, n)) for n, _node in enumerate(node))
    else:
        _iterator = [(node, value_range)]

    for _node, _value_range in _iterator:
        ln = LookupNode(loc=_value_range)
        ln.intelligence_node = get_source_completer(_node)
        code_intel.add_lookup_node(ln)


class CWLFieldType(CWLBaseType):

    def __init__(self, doc: str, required: bool, allowed_types: list):
        super().__init__("a field")
        self.doc = doc
        self.required = required
        if not isinstance(allowed_types, list):
            allowed_types = [allowed_types]
        self.types = allowed_types


def _put_this_field_first(_field_iterator, field_name):
    field_iterator = []
    for k, v in _field_iterator:
        if k == field_name:
            field_iterator = [(k, v)] + field_iterator
        else:
            field_iterator += [(k, v)]
    return field_iterator



"""
Copyright (c) 2016 Sourcegraph
Copyright (c) 2019 Seven Bridges. See LICENSE

This code is assembled and reorganized
Retrieved from: https://github.com/sourcegraph/python-langserver/blob/master/langserver/jsonrpc.py

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json
import logging
import queue
import threading
from collections import deque

logger = logging.getLogger(__name__)


class JSONRPC2ProtocolError(Exception):
    pass


class ReadWriter:
    def __init__(self, reader, writer):
        self.reader = reader
        self.writer = writer

    def readline(self, *args):
        return self.reader.readline(*args).decode("utf-8")

    def read(self, *args):
        return self.reader.read(*args).decode("utf-8")

    def write(self, out):
        self.writer.write(out.encode("utf-8"))
        self.writer.flush()


class TCPReadWriter(ReadWriter):
    def readline(self, *args):
        data = self.reader.readline(*args)
        return data.decode("utf-8")

    def read(self, *args):
        return self.reader.read(*args).decode("utf-8")

    def write(self, out):
        self.writer.write(out.encode())
        self.writer.flush()


class JSONRPC2Connection:
    def __init__(self, conn=None):
        self.conn = conn
        self._msg_buffer = deque()
        self._next_id = 1

    def _read_header_content_length(self, line):
        if len(line) < 2 or line[-2:] != "\r\n":
            raise JSONRPC2ProtocolError("Line endings must be \\r\\n")
        if line.startswith("Content-Length: "):
            _, value = line.split("Content-Length: ")
            value = value.strip()
            try:
                return int(value)
            except ValueError:
                raise JSONRPC2ProtocolError(
                    "Invalid Content-Length header: {}".format(value))

    def _receive(self):
        line = self.conn.readline()
        if line == "":
            raise EOFError()
        length = self._read_header_content_length(line)
        # Keep reading headers until we find the sentinel line for the JSON
        # request.
        while line != "\r\n":
            line = self.conn.readline()
        body = self.conn.read(length)
        logger.debug("RECV %s", body)
        return json.loads(body)

    def read_message(self, want=None):
        """Read a JSON RPC message sent over the current connection.
        If id is None, the next available message is returned.
        """
        if want is None:
            if self._msg_buffer:
                return self._msg_buffer.popleft()
            return self._receive()

        # First check if our buffer contains something we want.
        msg = deque_find_and_pop(self._msg_buffer, want)
        if msg:
            return msg

        # We need to keep receiving until we find something we want.
        # Things we don't want are put into the buffer for future callers.
        while True:
            msg = self._receive()
            if want(msg):
                return msg
            self._msg_buffer.append(msg)

    def _send(self, body):
        body = json.dumps(body, separators=(",", ":"))
        content_length = len(body)
        response = (
            "Content-Length: {}\r\n"
            "Content-Type: application/vscode-jsonrpc; charset=utf8\r\n\r\n"
            "{}".format(content_length, body))
        self.conn.write(response)
        logger.debug("SEND %s", body)

    def write_response(self, rid, result):
        body = {
            "jsonrpc": "2.0",
            "id": rid,
            "result": result,
        }
        self._send(body)

    def write_error(self, rid, code, message, data=None):
        e = {
            "code": code,
            "message": message,
        }
        if data is not None:
            e["data"] = data
        body = {
            "jsonrpc": "2.0",
            "id": rid,
            "error": e,
        }
        self._send(body)

    def send_request(self, method: str, params):
        rid = self._next_id
        self._next_id += 1
        body = {
            "jsonrpc": "2.0",
            "id": rid,
            "method": method,
            "params": params,
        }
        self._send(body)
        return self.read_message(want=lambda msg: msg.get("id") == rid)

    def send_notification(self, method: str, params):
        body = {
            "jsonrpc": "2.0",
            "method": method,
            "params": params,
        }
        self._send(body)

    def send_request_batch(self, requests):
        """Pipelines requests and returns responses.
        The responses is a generator where the nth response corresponds
        with the nth request. Users must read the generator until the
        end, otherwise you will leak a thread.
        """

        # We communicate the request ids using a thread safe queue.
        # It also allows us to bound the number of concurrent requests.
        q = queue.Queue(100)

        def send():
            for method, params in requests:
                rid = self._next_id
                self._next_id += 1
                q.put(rid)
                body = {
                    "jsonrpc": "2.0",
                    "id": rid,
                    "method": method,
                    "params": params,
                }
                self._send(body)

            # Sentinel value to indicate we are done
            q.put(None)

        threading.Thread(target=send).start()

        while True:
            rid = q.get()
            if rid is None:
                break
            yield self.read_message(want=lambda msg: msg.get("id") == rid)


def deque_find_and_pop(d, f):
    idx = None
    for i, v in enumerate(d):
        if f(v):
            idx = i
            break
    if idx is None:
        return None
    d.rotate(-idx)
    v = d.popleft()
    d.rotate(idx)
    return v



"""This keeps track of any context needed for auto-completions and expression evaluations
This therefore carries """

#  Copyright (c) 2019 Seven Bridges. See LICENSE

from typing import List
from dataclasses import dataclass

from .intelligence import IntelligenceNode
from .workflow import Workflow, WFStepIntelligence


@dataclass
class IntelligenceContext:
    path: List[str] = None
    workflow: Workflow = None
    workflow_step_intelligence: WFStepIntelligence = None
    requirements: IntelligenceNode = None


def copy_context(old: IntelligenceContext) -> IntelligenceContext:
    return IntelligenceContext(
        path=list(old.path),
        workflow=old.workflow,
        workflow_step_intelligence=old.workflow_step_intelligence,
        requirements=old.requirements)



"""
Generate inputs, outputs for processes, including intermediate outputs for workflows

The sample data is stored in a dictionary with the following structure

{
    "inputs":
        "in1": ...,
        "step1/out1": ...
    "outputs":
        "out1": ...,
        "out2": ...
}
"""

#  Copyright (c) 2019 Seven Bridges. See LICENSE

import random
import string

from ..cwl.lib import resolve_file_path, list_as_map
from .schemadef import extract_schemadef
from .yaml import fast_load


def get_sample_runtime(cwl: dict, doc_path: tuple):
    runtime = {
        "outdir": "/out/dir",
        "tmpdir": "/tmp/dir",
        "cores": 4,
        "ram": 1024,
        "outdirSize": 2048,
        "tmpdirSize": 4096
    }

    # https://www.commonwl.org/v1.1/CommandLineTool.html#Execution
    if "outputEval" in doc_path:
        try:
            if float(cwl.get("cwlVersion")[1:]) > 1.0:
                runtime["exitCode"] = 1
        except ValueError as e:
            pass

    return runtime


def get_sample_data(doc_uri: str, cwl: dict, user_types: dict):
    return {
        "inputs": {
            **generate_sample_inputs(cwl, user_types),
            **generate_sample_intermediate_results(doc_uri, cwl, user_types)
        },
        "outputs": generate_sample_outputs(cwl, user_types)
    }


def get_sample_globbed_files(name):
    return [
        basic_example_value(name + "/globbed_file_" + str(i), "File")
        for i in range(random.randint(0, 4))
    ]


def generate_sample_inputs(cwl: dict, user_types: dict):
    return generate_values(cwl.get("inputs"), user_types)


def generate_sample_outputs(cwl: dict, user_types: dict):
    return generate_values(cwl.get("outputs"), user_types)


def generate_sample_intermediate_results(doc_uri: str, cwl: dict, parent_user_types):
    return {
        step_id + "/" + k: v
        for step_id, step in list_as_map(cwl.get("steps"), key_field="id", problems=[]).items()
        for k, v in extract_step_sample_outputs(doc_uri, step.get("run"), parent_user_types).items()
    }


def generate_values(_ports: dict, user_types: dict):
    if not isinstance(_ports, (list, dict)):
        _ports = {}
    return {
        k: example_value(k, _type_v, user_types)
        for k, _type_v in list_as_map(_ports, key_field="id", problems=[]).items()
    }


# This should be invoked when we arrive at the "run" field of a workflow
def extract_step_sample_outputs(doc_uri: str, run_field, parent_user_types):

    # todo: verify this works with inlined steps
    user_types = parent_user_types
    if isinstance(run_field, str):
        linked_file = resolve_file_path(doc_uri, run_field)
        if linked_file.exists() and linked_file.is_file():
            run_field = fast_load.load(linked_file)
            user_types = extract_schemadef(linked_file.as_uri(), run_field)

    outputs = {}
    if isinstance(run_field, dict):
        outputs = generate_sample_outputs(run_field, user_types)

    return outputs


def basic_example_value(name, _type):
    name = name + "_" + "".join(random.choices(string.ascii_letters, k=5))
    if _type == 'null':
        return 'null'
    elif _type == 'Any':
        return 'Any'
    elif _type == 'boolean':
        return random.randint(0, 1) > 0
    elif _type == 'int' or _type == 'long':
        return random.randint(-1000, 1000)
    elif _type == 'float' or _type == 'double':
        return random.random() * 100 - 50
    elif _type == 'string':
        return name
    elif _type == 'File':
        return file_example_value(name, _type)
    elif _type == 'Directory':
        return {
            'class': 'Directory',
            'path': f'/path/to/{name}'
        }


def file_example_value(name, cwl_type, ext=".ext"):
    ex_file = _example_file(name, ext)
    if isinstance(cwl_type, dict) and "secondaryFiles" in cwl_type:
        ex_file["secondaryFiles"] = [
            _example_file(name, sec_ext)
            for sec_ext in cwl_type.get("secondaryFiles")
        ]
    return ex_file


def _example_file(name, ext):
    fsize = random.randint(0, 100)
    contents = "".join(random.choices(string.ascii_letters, k=fsize))
    return {
        'class': 'File',
        'path': f'/path/to/{name}.ext',
        'location': f'/location/of/{name}.ext',
        'basename': f'{name}.ext',
        'dirname': '/path/to/',
        'nameroot': name,
        'nameext': '.ext',
        'checksum': "sha1$deadbeef",
        'size': fsize,
        'format': 'someformat',
        'contents': contents
    }


def enum_example_value(symbols):
    return symbols[random.randint(0, len(symbols) - 1)]


def record_example_value(name, _type, user_types):
    return {
        k: example_value(f"{name}/{k}", _type_v, user_types)
        for k, _type_v in list_as_map(_type.get("fields"), key_field="name", problems=[]).items()
    }


def example_value(name, cwl_type, user_types, array_of=False):
    if array_of:
        return [example_value(name + "/" + str(i), cwl_type, user_types) for i in range(4)]

    if isinstance(cwl_type, list):
        l = len(cwl_type)
        return example_value(name, cwl_type[random.randint(0, l - 1)], user_types)

    elif isinstance(cwl_type, dict) and "type" in cwl_type:
        _type = cwl_type.get("type")
        if _type == "array":
            return example_value(name, cwl_type.get("items"), user_types, array_of=True)
        elif _type == "enum":
            return enum_example_value(cwl_type.get("symbols"))
        elif _type == "record":
            return record_example_value(name, cwl_type, user_types)
        elif _type == "File":
            return file_example_value(name, cwl_type)
        else:
            return example_value(name, _type, user_types)

    elif isinstance(cwl_type, str):
        # desugar
        if cwl_type.endswith("?"):
            cwl_type = cwl_type[:-1]

        if cwl_type.endswith("[]"):
            cwl_type = {
                "type": "array",
                "items": [cwl_type[:-2]]
            }
            return example_value(name, cwl_type, user_types)

        if cwl_type in user_types:
            return example_value(name, user_types.get(cwl_type), user_types)

        return basic_example_value(name, cwl_type)



#  Copyright (c) 2019 Seven Bridges. See LICENSE

from .basetype import CWLBaseType, IntelligenceContext, Intelligence, MapSubjectPredicate, TypeCheck, Match
from ..langserver.lspobjects import Range, CompletionItem, Diagnostic, DiagnosticSeverity, Hover
from ..code.intelligence import LookupNode
from ..code.yaml import yaml_to_string

import logging
logger = logging.getLogger(__name__)


class CWLEnumType(CWLBaseType):

    def __init__(self, name: str, symbols: set):
        super().__init__(name)
        self.symbols = symbols
        self._hover_value = None

    def check(self, node, node_key: str=None, map_sp: MapSubjectPredicate=None) -> TypeCheck:

        if not (isinstance(node, str) or None):
            return TypeCheck(
                cwl_type=self,
                match=Match.No)
        else:
            if self.name in ["PrimitiveType", "CWLType"]:
                # Special treatment for data types
                return TypeCheck(cwl_type=CWLDataType(node, self.symbols))
            else:
                return TypeCheck(cwl_type=self)

    def parse(self,
              doc_uri: str,
              node,
              intel_context: IntelligenceContext,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):

        if node not in self.symbols:
            problems += [
                Diagnostic(
                    _range=value_range,
                    message=f"Expecting one of: {sorted(self.symbols)}",
                    severity=DiagnosticSeverity.Error)
            ]

        ln = LookupNode(loc=value_range)
        ln.intelligence_node = self
        code_intel.add_lookup_node(ln)

    def completion(self):
        return [CompletionItem(label=s) for s in self.symbols]


class CWLDataType(CWLEnumType):

    def parse(self,
              doc_uri: str,
              node,
              intel_context: IntelligenceContext,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):
        # add in user defined data types
        self.symbols = set(code_intel.type_defs.keys()).union(self.symbols)

        # de-sugar
        if node[-1] == "?":
            node = node[:-1]
        if node[-2:] == "[]":
            node = node[:-2]

        if node not in self.symbols:
            problems += [
                Diagnostic(
                    _range=value_range,
                    message=f"Expecting one of: {sorted(self.symbols)}",
                    severity=DiagnosticSeverity.Error)
            ]

        elif node in code_intel.type_defs:
            self._hover_value = code_intel.type_defs[node]

        ln = LookupNode(loc=value_range)
        ln.intelligence_node = self
        code_intel.add_lookup_node(ln)

    def hover(self):
        if self._hover_value is not None:
            return Hover(yaml_to_string(self._hover_value))




#  Copyright (c) 2019 Seven Bridges. See LICENSE

from .basetype import CWLBaseType, Intelligence, MapSubjectPredicate
from ..langserver.lspobjects import Range, CompletionItem
from ..code.intelligence import LookupNode
from ..code.intelligencecontext import IntelligenceContext

import logging
logger = logging.getLogger(__name__)


class CWLRequirementsType(CWLBaseType):

    def __init__(self, name, req_types):
        super().__init__(name)
        self.req_types = req_types

    def parse(self,
              doc_uri: str,
              node,
              intel_context: IntelligenceContext,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):

        ln = LookupNode(loc=value_range)
        ln.intelligence_node = self
        code_intel.add_lookup_node(ln)

    def completion(self):
        return [CompletionItem(label=k.name) for k in self.req_types]



"""Parse CWL and create a JSON file describing the workflow. This dictionary
is directly suitable for display by vis.js, but can be parsed for any other
purpose."""

#  Copyright (c) 2019 Seven Bridges. See LICENSE

from ..cwl.lib import ListOrMap


def cwl_graph(cwl: dict):

    graph = {
        "nodes": [],
        "edges": [],
        "line-numbers": {}
    }

    inputs = ListOrMap(cwl.get("inputs", {}), key_field="id", problems=[])
    _add_nodes(graph, inputs, "inputs")

    steps = ListOrMap(cwl.get("steps", {}), key_field="id", problems=[])
    _add_nodes(graph, steps, "steps")

    outputs = ListOrMap(cwl.get("outputs", {}), key_field="id", problems=[])
    _add_nodes(graph, outputs, "outputs")

    _add_edges(graph, inputs, outputs, steps)

    return graph


def _add_nodes(graph, grp, grp_id):
    for k, v in grp.as_dict.items():
        graph["nodes"] += [{
            "id": k,
            "label": v.get("label", k) if isinstance(v, dict) else k,
            "group": grp_id
        }]
        _mark_step_modifiers(graph["nodes"][-1], v)
        graph["line-numbers"][k] = grp.get_range_for_value(k).start.line


def _mark_step_modifiers(node_data, node):
    if not isinstance(node, dict):
        return

    code, title = "", []
    _scatter = node.get("scatter")
    _conditional = node.get("when")
    if _scatter:
        code += "S"
        title += [f"Scatter on {_scatter}"]
    if _conditional:
        code += "?"
        title += [f"Conditional on {_conditional}"]
    if code:
        node_data["icon"] = {
            "code": code,
            "color": "#ffffff"
        }
    if title:
        node_data["title"] = "<br/>".join(title)


def _add_edges(graph, inputs, outputs, steps):

    for k, v in steps.as_dict.items():
        _to = k
        for _, prt in ListOrMap(v.get("in", {}), key_field="id", problems=[]).as_dict.items():
            graph["edges"] += [{"from": _f, "to": _to} for _f in _get_source_step(prt, "source")]

    for k, v in outputs.as_dict.items():
        _to = k
        graph["edges"] += [{"from": _f, "to": _to} for _f in _get_source_step(v, "outputSource")]


def _get_source_step(v, key):
    src = v.get(key) if isinstance(v, dict) else v
    if not isinstance(src, list):
        src = [src]
    return [s.split("/")[0] for s in src if isinstance(s, str)]



"""
https://microsoft.github.io/language-server-protocol/specification#textDocument_didOpen

The document open notification is sent from the client to the server to signal
newly opened text documents. The document’s truth is now managed by the client
and the server must not try to read the document’s truth using the document’s Uri.
Open in this sense means it is managed by the client. It doesn’t necessarily
mean that its content is presented in an editor. An open notification must
not be sent more than once without a corresponding close notification send
before. This means open and close notification must be balanced and the max
open count for a particular textDocument is one. Note that a server’s ability
to fulfill requests is independent of whether a text document is open or closed.

Notes:
Multiple documents can be open at the same time. We need to keep track of this.

interface TextDocumentItem {
    /**
     * The text document's URI.
     */
    uri: DocumentUri;

    /**
     * The text document's language identifier.
     */
    languageId: string;

    /**
     * The version number of this document (it will increase after each
     * change, including undo/redo).
     */
    version: number;

    /**
     * The content of the opened text document.
     */
    text: string;
}
"""
#  Copyright (c) 2019 Seven Bridges. See LICENSE

from .lspobjects import to_dict, PublishDiagnosticsParams
from .base import CWLLangServerBase
from ..code.document import Document

import logging
logger = logging.getLogger(__name__)


class FileOperation(CWLLangServerBase):

    def serve_textDocument_didOpen(self, client_query):
        params = client_query["params"]
        doc_uri = params["textDocument"]["uri"]

        document = Document(
            doc_uri=doc_uri,
            scratch_path=self.config.scratch_path,
            text=params["textDocument"]["text"],
            version=params["textDocument"]["version"],
            type_dicts=self.config.lang_models)

        self.open_documents[doc_uri] = document
        self._mark_document_issues(doc_uri)

    def serve_textDocument_didChange(self, client_query):
        params = client_query["params"]
        doc_uri = params["textDocument"]["uri"]

        content_changes = params["contentChanges"]

        if len(content_changes) > 1:
            logger.error("Server can currently only handle full text updates")

        content_change = content_changes[0]
        if "range" in content_change or "rangeLength" in content_change:
            logger.error("Server can currently only handle full text updates")

        self.open_documents[doc_uri].update(new_text=content_change["text"])
        self._mark_document_issues(doc_uri)

    def serve_textDocument_didClose(self, client_query):
        params = client_query["params"]
        doc_uri = params["textDocument"]["uri"]
        self.open_documents.pop(doc_uri)

    def _mark_document_issues(self, doc_uri):
        document = self.open_documents[doc_uri]
        self.conn.send_notification(
            method="textDocument/publishDiagnostics",
            params=to_dict(
                PublishDiagnosticsParams(
                    uri=doc_uri,
                    diagnostics=document.problems)))



"""Manages aspects related to test executions of the CWL and of JS expressions."""

#  Copyright (c) 2019 Seven Bridges. See LICENSE

import pathlib

from ..cwl.lib import un_mangle_uri, list_as_map
from .sampledata import (
    get_sample_runtime,
    get_sample_data,
    get_sample_globbed_files)

from ruamel.yaml import YAML

import logging
logger = logging.getLogger(__name__)

fast_yaml_io = YAML(typ='safe')
fast_yaml_io.default_flow_style = False

job_inputs_ext = ".benten.test.job.yml"


class ExecutionContext:
    """Carries the job object (sample inputs), expression lib and, if a workflow, simulated
    outputs for each step"""

    def __init__(self, doc_uri: str, cwl: dict, user_types: dict, scratch_path: pathlib.Path):
        self.doc_uri = doc_uri
        self.cwl = cwl
        self.user_types = user_types
        self.scratch_path = scratch_path
        self.expression_lib = []
        self._sample_data = None
        # self._intermediate_outputs = None

    def runtime(self, doc_path: tuple):
        return get_sample_runtime(self.cwl, doc_path)

    @property
    def sample_data(self):
        ex_job_file = self.get_sample_data_file_path()
        if ex_job_file.exists():
            if ex_job_file.open().readline().startswith("#custom"):
                self._sample_data = fast_yaml_io.load(ex_job_file.open().read() or "")
                return self._sample_data

        if self._sample_data is None:
            self._sample_data = get_sample_data(self.doc_uri, self.cwl, self.user_types)
            ex_job_file.parent.mkdir(parents=True, exist_ok=True)
            fast_yaml_io.dump(self._sample_data, ex_job_file)

        return self._sample_data

    def get_workflow_step_inputs(self, doc_path: tuple):
        step_id = doc_path[1]
        step_sample_outputs = self.sample_data["inputs"]
        step_obj = list_as_map(self.cwl.get("steps"), key_field="id", problems=[]).get(step_id)
        input_obj = {}
        for k, in_obj in list_as_map(step_obj.get("in"), key_field="id", problems=[]).items():
            if isinstance(in_obj, dict):
                src = in_obj.get("source")
            else:
                src = in_obj

            if isinstance(src, list):
                input_obj[k] = [step_sample_outputs.get(s) for s in src]
            else:
                input_obj[k] = step_sample_outputs.get(src)

        self_obj = input_obj.get(doc_path[3])

        return input_obj, self_obj

    @staticmethod
    def get_sample_globbed_files(name):
        return get_sample_globbed_files(name)

    def set_expression_lib(self, expression_lib: list=None):
        self.expression_lib = expression_lib

    def get_sample_data_file_path(self) -> pathlib.Path:
        return self.scratch_path / pathlib.Path(
            *un_mangle_uri(self.doc_uri).with_suffix(job_inputs_ext).parts[1:])



"""Short script that uses schema-salad to generate and post-process the CWL yml document into
a JSON format that is read in by Benten and converted into a set of Python classes for CWL types"""
#  Copyright (c) 2019 Seven Bridges. See LICENSE

import sys
import subprocess
import json


def remove_uris(d):
    if isinstance(d, dict):
        for k, v in d.items():
            d[k] = remove_uris(v)
        return d
    elif isinstance(d, list):
        return [remove_uris(v) for v in d]
    else:
        if isinstance(d, str):
            if d.startswith("https://w3id.org") or d.startswith("http://www.w3.org"):
                _v = d.split("#")[1]
                if "/" in _v:
                    _v = _v.split("/")[1]
                return _v
        return d


def main():
    print(sys.argv)
    if len(sys.argv) != 3:
        print("python create-schema.py <input.yml> <output.json>\n"
              "e.g. python create-schema.py CommonWorkflowLanguage.yml schema-v1.0.json")
        return

    #subprocess.call(f"schema-salad-tool --print-pre {sys.argv[1]} > {sys.argv[2]}", shell=True)
    subprocess.call(f"schema-salad-tool --print-avro {sys.argv[1]} > {sys.argv[2]}", shell=True)
    with open(sys.argv[2], "r") as f:
        d = json.load(f)
        d = remove_uris(d)

    with open(sys.argv[2], "w") as f:
        json.dump(d, f, indent=2)


if __name__ == "__main__":
    main()



#  Copyright (c) 2019 Seven Bridges. See LICENSE

import pathlib
import urllib.parse
import urllib.request
import urllib.error

from ..langserver.lspobjects import Diagnostic, DiagnosticSeverity, Range, Position
from ..code.yaml import fast_yaml_load


def get_range_for_key(parent, key):
    start = parent.lc.key(key)
    end = (start[0], start[1] + len(key))
    return Range(Position(*start), Position(*end))


# TODO: refactor this for redundancy
def get_range_for_value(node, key):
    if isinstance(node, dict):
        start = node.lc.value(key)
    else:
        start = node.lc.item(key)

    v = node[key]
    if v is None:
        end = (start[0], start[1])
    else:
        v = str(v)
        _lines = v.splitlines() or [""]
        ln_cnt = len(_lines)
        if ln_cnt == 1:
            ln_cnt -= 1
        # Multi-line strings begin from the next line as opposed to single line strings

        # Ideally we'd be carrying around the raw lines to compute the end of the last line
        # but this is good enough and less cumbersome
        end = (start[0] + ln_cnt, (len(_lines[-1]) + start[1]) if ln_cnt == 0 else 1000)

    return Range(Position(*start), Position(*end))


class ListOrMap:

    def __init__(self, node, key_field, problems):
        self.was_dict = None
        self.as_dict = {}
        self.original_obj = node
        self.key_ids = {}
        self.map_key_to_idx = {}
        if isinstance(node, dict):
            self.as_dict = node
            self.was_dict = True
        elif isinstance(node, list):
            self.was_dict = False
            for n, _item in enumerate(node):
                if isinstance(_item, dict):
                    key = _item.get(key_field)
                    if key is not None:
                        self.as_dict[key] = _item
                        self.key_ids[key] = get_range_for_value(_item, key_field)
                        self.map_key_to_idx[key] = n
                    else:
                        problems += [
                            Diagnostic(
                                _range=get_range_for_value(node, n),
                                message=f"Missing key field {key_field}",
                                severity=DiagnosticSeverity.Error)
                        ]

    def get_range_for_id(self, key):
        if self.was_dict:
            return get_range_for_key(self.as_dict, key)
        else:
            return self.key_ids[key]

    def get_range_for_value(self, key):
        if self.was_dict:
            return get_range_for_value(self.as_dict, key)
        else:
            return get_range_for_value(self.original_obj, self.map_key_to_idx[key])


# TODO: Deprecate this function
def list_as_map(node, key_field, problems):
    if isinstance(node, dict):
        return node

    new_node = {}

    if isinstance(node, list):
        for n, _item in enumerate(node):
            if isinstance(_item, dict):
                key = _item.get(key_field)
                if key is not None:
                    new_node[key] = _item
                else:
                    problems += [
                        Diagnostic(
                            _range=get_range_for_value(node, n),
                            message=f"Missing key field {key_field}",
                            severity=DiagnosticSeverity.Error)
                    ]

    return new_node


def validate_and_load_linked_file(doc_uri: str, path: str, loc: Range, problems: list) -> (str, str, dict):

    full_path, contents, node_dict = path, "", {}

    if urllib.parse.urlparse(path).scheme != "":
        try:
            contents = urllib.request.urlopen(path).read().decode('utf-8')
            node_dict = fast_yaml_load(contents)
        except urllib.error.HTTPError:
            problems += [
                Diagnostic(
                    _range=loc,
                    message=f"Missing URL: {path}",
                    severity=DiagnosticSeverity.Error)
            ]

        return full_path, contents, node_dict

    linked_file = resolve_file_path(doc_uri, path)
    if not linked_file.exists():
        problems += [
            Diagnostic(
                _range=loc,
                message=f"Missing document: {path}",
                severity=DiagnosticSeverity.Error)
        ]
    elif not linked_file.is_file():
        problems += [
            Diagnostic(
                _range=loc,
                message=f"Linked document must be file: {path}",
                severity=DiagnosticSeverity.Error)
        ]
    else:
        contents = linked_file.open("r").read()
        node_dict = fast_yaml_load(contents)

    return linked_file, contents, node_dict


# doc_uri on windows takes the form "file:///c%3A/Users/me/a.cwl"
# parsing the uri into components results in path being "/c%3A/Users/me/a.cwl"
# This needs to be
# a) decoded (so that c%3A -> c:)
# b) The leading "/" needs to be chomped.
# Step a) is redundant on *nix and b) should not be done
def un_mangle_uri(doc_uri):
    _my_path = pathlib.Path(urllib.parse.unquote(urllib.parse.urlparse(doc_uri).path))
    if isinstance(_my_path, pathlib.WindowsPath):
        _my_path = pathlib.Path(str(_my_path)[1:])
    return _my_path


def resolve_file_path(doc_uri, target_path):
    _path = pathlib.PurePosixPath(target_path)
    if not _path.is_absolute():
        base_path = un_mangle_uri(doc_uri).parent
    else:
        base_path = "."
    _path = pathlib.Path(base_path / _path).resolve().absolute()
    return _path



"""Collection of common LSP objects"""
#  Copyright (c) 2019 Seven Bridges. See LICENSE

from typing import List
from enum import IntEnum


def to_dict(v):
    if isinstance(v, LSPObject):
        return {
            k: to_dict(_v)
            for k, _v in v.__dict__.items() if _v is not None
        }
    elif isinstance(v, dict):
        return {
            k: to_dict(_v)
            for k, _v in v.items()
        }
    elif isinstance(v, list):
        return [
            to_dict(_v) for _v in v
        ]
    else:
        return v


class LSPObject:
    def to_dict(self):
        return to_dict(self)


class Position(LSPObject):
    def __init__(self, line, character):
        self.line = line
        self.character = character

    def __hash__(self):
        return hash((self.line, self.character))

    def __eq__(self, other):
        if not isinstance(other, type(self)): return NotImplemented
        return self.line == other.line and self.character == other.character


class Range(LSPObject):
    def __init__(self, start: Position, end: Position):
        self.start = start
        self.end = end

    def __hash__(self):
        return hash((self.start, self.end))

    def __eq__(self, other):
        if not isinstance(other, type(self)): return NotImplemented
        return self.start == other.start and self.end == other.end


class TextEdit(LSPObject):
    def __init__(self, _range: Range, new_text: str):
        self.range = _range
        self.newText = new_text


class Location(LSPObject):
    def __init__(self, uri, _range: Range=Range(Position(0, 0), Position(0, 0))):
        self.uri = uri
        self.range = _range


class DiagnosticSeverity(IntEnum):
    Error = 1
    Warning = 2
    Information = 3
    Hint = 4


class Diagnostic(LSPObject):
    def __init__(self,
                 _range: Range, message: str,
                 severity: DiagnosticSeverity=None,
                 code: str=None,
                 source: str=None):
        self.range = _range
        self.message = message
        self.severity = severity
        self.code = code
        self.source = source

    def __hash__(self):
        return hash((self.range, self.message))

    def __eq__(self, other):
        if not isinstance(other, type(self)): return NotImplemented
        return self.range == other.range and self.message == other.message


class PublishDiagnosticsParams(LSPObject):
    def __init__(self, uri, diagnostics: List[Diagnostic]):
        self.uri = uri
        self.diagnostics = diagnostics


class CodeActionKind:
    QuickFix = 'quickfix'
    CodeActionKind = 'refactor'
    RefactorExtract = 'refactor.extract'
    RefactorInline = 'refactor.inline'
    RefactorRewrite = 'refactor.rewrite'
    Source = 'source'
    SourceOrganizeImports = 'source.organizeImports'


class CompletionItemKind(IntEnum):
    Text = 1
    Method = 2
    Function = 3
    Constructor = 4
    Field = 5
    Variable = 6
    Class = 7
    Interface = 8
    Module = 9
    Property = 10
    Unit = 11
    Value = 12
    Enum = 13
    Keyword = 14
    Snippet = 15
    Color = 16
    File = 17
    Reference = 18
    Folder = 19
    EnumMember = 20
    Constant = 21
    Struct = 22
    Event = 23
    Operator = 24
    TypeParameter = 25


class InsertTextFormat(IntEnum):
    PlainText = 1
    Snippet = 2


class CompletionItem(LSPObject):
    def __init__(self,
                 label: str,
                 text_edit: TextEdit = None, additional_text_edits: [TextEdit] = None,
                 kind: CompletionItemKind = CompletionItemKind.Text,
                 insert_text_format: InsertTextFormat = None,
                 detail: str = None, documentation: str = None, preselect: bool = None,
                 sort_text: str = None, filter_text: str = None):
        self.label = label
        self.kind = kind
        self.detail = detail
        self.documentation = documentation
        self.preselect = preselect
        self.sortText = sort_text
        self.filterText = filter_text
        self.insertTextFormat = insert_text_format
        self.textEdit = text_edit
        self.additionalTextEdits = additional_text_edits

    def set_range(self, _range: Range):
        self.textEdit.range = _range

    @classmethod
    def from_snippet(
            cls,
            snippet: dict):

        if "label" not in snippet:
            snippet["label"] = "Unknown"

        try:
            snippet["kind"] = CompletionItemKind[snippet["kind"]]
        except KeyError:
            snippet["kind"] = CompletionItemKind.Text

        snippet["insert_text_format"] = InsertTextFormat.Snippet
        # snippet["insert_text"] = snippet.get("snippet")
        snippet["text_edit"] = TextEdit(
            _range=Range(
                start=Position(line=0, character=0),
                end=Position(line=0, character=0)
            ),
            new_text=snippet.get("text_edit"))

        return cls(**snippet)


class CompletionList(LSPObject):
    def __init__(self, is_incomplete: bool = False, items: [CompletionItem] = None):
        self.isIncomplete = is_incomplete
        self.items = items or []


class SymbolKind(IntEnum):
    File = 1
    Module = 2
    Namespace = 3
    Package = 4
    Class = 5
    Method = 6
    Property = 7
    Field = 8
    Constructor = 9
    Enum = 10
    Interface = 11
    Function = 12
    Variable = 13
    Constant = 14
    String = 15
    Number = 16
    Boolean = 17
    Array = 18
    Object = 19
    Key = 20
    Null = 21
    EnumMember = 22
    Struct = 23
    Event = 24
    Operator = 25
    TypeParameter = 26


class DocumentSymbol(LSPObject):
    def __init__(self, name, kind, _range, selection_range, detail=None, children=None):
        self.name = name
        self.detail = detail
        self.kind: SymbolKind = kind
        self.range: Range = _range
        self.selectionRange: Range = selection_range
        self.children: List[DocumentSymbol] = children


class Hover(LSPObject):
    def __init__(self, contents, _range=None, wrap_as_code_block=True):
        if wrap_as_code_block:
            self.contents = MarkupContent(
                kind="markdown",
                value="```\n" + contents + "\n```"
            )
        else:
            self.contents = contents

        self.range = _range


class MarkupContent(LSPObject):
    def __init__(self, value, kind="markdown"):
        self.kind = kind
        self.value = value



#  Copyright (c) 2019 Seven Bridges. See LICENSE

__all__ = [
    'CWLBaseType', 'MapSubjectPredicate', 'TypeCheck', 'Match',
    'CWLUnknownType', 'CWLAnyType', 'CWLExpressionType',
    'CWLEnumType', 'CWLArrayType', 'CWLListOrMapType', 'CWLRecordType', 'CWLFieldType'
]

from .basetype import CWLBaseType, MapSubjectPredicate, TypeCheck, Match
from .unknowntype import CWLUnknownType
from .anytype import CWLAnyType
from .expressiontype import CWLExpressionType
from .enumtype import CWLEnumType
from .arraytype import CWLArrayType
from .lomtype import CWLListOrMapType
from .recordtype import CWLRecordType, CWLFieldType



#  Copyright (c) 2019 Seven Bridges. See LICENSE

from .basetype import CWLBaseType, MapSubjectPredicate
from ..code.intelligence import Intelligence
from ..code.intelligencecontext import IntelligenceContext
from ..langserver.lspobjects import Range, Diagnostic, DiagnosticSeverity

import logging
logger = logging.getLogger(__name__)


class CWLNameSpacedType(CWLBaseType):

    def __init__(self, name):
        super().__init__(name)

    def parse(self,
              doc_uri: str,
              node,
              intel_context: IntelligenceContext,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):

        _s = self.name.split(":")
        if _s[0] not in code_intel.namespaces:
            problems += [
                Diagnostic(
                    _range=key_range or value_range,
                    message=f"Expecting one of: {sorted(code_intel.namespaces)}",
                    severity=DiagnosticSeverity.Error)
            ]



"""Code to parse step interfaces and connectivity and present results in
a generic way that can be used by code intelligence as well as graph
drawing routines.

We try and be parsimonious about our parsing. On the first pass through,
when we analyze all CWL types, we parse all the individual steps with
this code. We keep references to all the steps in a separate
structure.

In the end we do a global analysis of the workflow, flagging connectivity
problems and building a graph of the workflow. We use this global analysis
to enable port completion. For all of this we reuse the previously extracted
step information.
"""

#  Copyright (c) 2019 Seven Bridges. See LICENSE

from typing import Dict

from ..cwl.lib import (get_range_for_value, list_as_map, ListOrMap)
from .intelligence import IntelligenceNode, CompletionItem
from ..langserver.lspobjects import Diagnostic, DiagnosticSeverity


import logging
logger = logging.getLogger(__name__)


# TODO: Use typed objects for ports to check types
class StepInterface:
    def __init__(self, inputs=None, outputs=None):
        self.inputs = inputs or set()
        self.outputs = outputs or set()


class Workflow:
    def __init__(self, inputs, outputs, steps):
        self._inputs = inputs
        self._outputs = outputs
        self._steps = steps

        self.step_intels: Dict[str, WFStepIntelligence] = {}
        self.wf_inputs = set(list_as_map(inputs, key_field="id", problems=[]).keys())
        self.wf_outputs = set(list_as_map(outputs, key_field="id", problems=[]).keys())

    def validate_connections(self, problems):
        unused_ports = set(self.wf_inputs)
        self.validate_step_connections(unused_ports, problems)
        self.validate_outputs(unused_ports, problems)
        self.flag_unused_inputs(unused_ports, problems)

    def validate_outputs(self, unused_ports, problems):
        outputs = ListOrMap(self._outputs, key_field="id", problems=[])
        for output_id, output in outputs.as_dict.items():
            _validate_source(
                port=output,
                src_key="outputSource",
                value_range=outputs.get_range_for_value(output_id),
                step_id=None,
                workflow=self,
                unused_ports=unused_ports,
                problems=problems)

    def validate_step_connections(self, unused_ports, problems):
        _steps = ListOrMap(self._steps, key_field="id", problems=[])
        for step_id, step in _steps.as_dict.items():
            step_intel = self.step_intels.get(step_id)
            if step_intel and isinstance(step, dict):
                step_intel.validate_connections(
                    ListOrMap(step.get("in"), key_field="id", problems=[]),
                    unused_ports=unused_ports,
                    problems=problems)

    def flag_unused_inputs(self, unused_ports, problems):
        inputs = ListOrMap(self._inputs, key_field="id", problems=[])
        for inp in unused_ports:
            if inp in inputs.as_dict:
                problems += [
                    Diagnostic(
                        _range=inputs.get_range_for_id(inp),
                        message=f"Unused input",
                        severity=DiagnosticSeverity.Warning)
                ]

    def add_step_intel(self, step_id, step_intel: 'WFStepIntelligence'):
        step_intel.workflow = self
        self.step_intels[step_id] = step_intel

    def get_step_intel(self, step_id):
        return self.step_intels.get(step_id)

    def get_output_source_completer(self, prefix):
        return WFOutputSourceCompleter(self, prefix)


class WFStepIntelligence:
    def __init__(self, step_id):
        super().__init__()
        self.step_id = step_id
        self.step_interface: StepInterface = StepInterface()
        self.workflow = None

    def set_step_interface(self, step_interface: StepInterface):
        self.step_interface = step_interface

    def validate_connections(self, inputs: ListOrMap, unused_ports, problems):
        if self.workflow is None:
            raise RuntimeError("Need to attach workflow first")

        for port_id, port in inputs.as_dict.items():
            if port_id not in self.step_interface.inputs:
                problems += [
                    Diagnostic(
                        _range=inputs.get_range_for_id(port_id),
                        message=
                        f"Expecting one of: {self.step_interface.inputs}"
                        if self.step_interface.inputs else
                        "No input ports found for this step",
                        severity=DiagnosticSeverity.Error)
                ]

            else:
                _validate_source(
                    port=port,
                    src_key="source",
                    value_range=inputs.get_range_for_value(port_id),
                    step_id=self.step_id,
                    workflow=self.workflow,
                    unused_ports=unused_ports,
                    problems=problems)

    def get_step_inport_completer(self):
        return WFStepInputPortCompleter(inputs=self.step_interface.inputs)

    def get_step_output_completer(self):
        return WFStepOutputPortCompleter(outputs=self.step_interface.outputs)

    def get_step_source_completer(self, prefix):
        return PortSourceCompleter(self, prefix)


class WFStepInputPortCompleter(IntelligenceNode):
    def __init__(self, inputs):
        super().__init__(completions=inputs)


class WFStepOutputPortCompleter(IntelligenceNode):
    def __init__(self, outputs):
        super().__init__(completions=outputs)


class PortSourceCompleterBase(IntelligenceNode):
    def __init__(self, prefix):
        super().__init__()
        self.prefix = prefix

    def _completion(self, workflow, step_id=None):
        if "/" not in self.prefix:
            return [
                CompletionItem(label=_id)
                for _port in (workflow.step_intels.keys(), workflow.wf_inputs)
                for _id in _port
                if _id != step_id
            ]
        else:
            src_step, src_port = self.prefix.split("/")
            step_intel = workflow.step_intels.get(src_step)
            if step_intel is not None:
                return [
                    CompletionItem(label=_id)
                    for _id in step_intel.step_interface.outputs
                ]


class PortSourceCompleter(PortSourceCompleterBase):
    def __init__(self, step_intel: WFStepIntelligence, prefix):
        super().__init__(prefix)
        self.step_intel = step_intel

    def completion(self):
        return self._completion(workflow=self.step_intel.workflow, step_id=self.step_intel.step_id)


class WFOutputSourceCompleter(PortSourceCompleterBase):
    def __init__(self, workflow: Workflow, prefix):
        super().__init__(prefix)
        self.workflow = workflow

    def completion(self):
        return self._completion(workflow=self.workflow)


# This should be invoked when we arrive at the "run" field of a workflow
def parse_step_interface(run_field: dict, problems: list):

    step_interface = StepInterface()

    if isinstance(run_field, dict):
        step_interface = StepInterface(
            inputs=set(list_as_map(run_field.get("inputs"),
                                   key_field="id",
                                   problems=problems).keys()),
            outputs=set(list_as_map(run_field.get("outputs"),
                                    key_field="id",
                                    problems=problems).keys()))

    return step_interface


def _validate_source(port, src_key, value_range, step_id, workflow, unused_ports, problems):

    src = None
    if isinstance(port, (str, list)):
        src = port
    elif isinstance(port, dict):
        if src_key in port:
            src = port.get(src_key)
            value_range = get_range_for_value(port, src_key)

    if src is None:
        return

    if isinstance(src, list):
        for n, _src in enumerate(src):
            _validate_one_source(_src, get_range_for_value(src, n), step_id, workflow, unused_ports, problems)
    elif isinstance(src, str):
        _validate_one_source(src, value_range, step_id, workflow, unused_ports, problems)


def _validate_one_source(src, value_range, step_id, workflow, unused_ports, problems):

    if src is None:
        return

    unused_ports.discard(src)

    if src in workflow.wf_inputs:
        return

    err_msg = f"No such workflow input. Expecting one of {workflow.wf_inputs}"

    if isinstance(src, str) and "/" in src:
        src_step, src_port = src.split("/")
        err_msg = "Port can not connect to same step"
        if src_step != step_id:
            err_msg = f"No step called {src_step}"
            if src_step in workflow.step_intels:
                err_msg = f"{src_step} has no port called {src_port}"
                if src_port in workflow.step_intels[src_step].step_interface.outputs:
                    return

    problems += [
        Diagnostic(
            _range=value_range,
            message=err_msg,
            severity=DiagnosticSeverity.Error)
    ]



#  Copyright (c) 2019 Seven Bridges. See LICENSE

from dataclasses import dataclass
from enum import IntEnum

from ..langserver.lspobjects import Range
from ..code.intelligence import IntelligenceNode, Intelligence
from ..code.intelligencecontext import IntelligenceContext


class MapSubjectPredicate:

    def __init__(self, subject, predicate):
        self.subject = subject
        self.predicate = predicate


class Match(IntEnum):
    Yes = 0
    Maybe = 1
    No = 2


@dataclass
class TypeCheck:
    cwl_type: 'CWLBaseType'
    match: Match = Match.Yes
    missing_req_fields: list = None
    missing_opt_fields: list = None


class CWLBaseType(IntelligenceNode):

    def __init__(self, name):
        super().__init__()
        self.name = name

    def check(self, node, node_key: str=None, map_sp: MapSubjectPredicate=None) -> TypeCheck:
        pass

    def parse(self,
              doc_uri: str,
              node,
              intel_context: IntelligenceContext,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):
        pass



#  Copyright (c) 2019 Seven Bridges. See LICENSE

import os
import shutil
from pathlib import Path as P
import configparser

from .cwl.specification import parse_schema

import logging
logger = logging.getLogger(__name__)


sbg_config_dir = P("sevenbridges", "benten")


# https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html

xdg_config_dir = {
    "env": "XDG_CONFIG_HOME",
    "default": P(P.home(), ".config")
}

# There is a raging debate on this and people want to add a new field to the XDG spec
# Me, I think logs are user data ...
xdg_data_home = {
    "env": "XDG_DATA_HOME",
    "default": P(P.home(), ".local", "share")
}

default_config_data_dir = P(P(__file__).parent, "000.package.data")


class Configuration(configparser.ConfigParser):
    def __init__(self):
        super().__init__()

        self.cfg_path = P(os.getenv(xdg_config_dir["env"], xdg_config_dir["default"]), sbg_config_dir)
        self.log_path = P(os.getenv(xdg_data_home["env"], xdg_data_home["default"]), sbg_config_dir, "logs")
        self.scratch_path = P(os.getenv(xdg_data_home["env"], xdg_data_home["default"]), sbg_config_dir, "scratch")

        if not self.log_path.exists():
            self.log_path.mkdir(parents=True)

        if not self.scratch_path.exists():
            self.scratch_path.mkdir(parents=True)

        self.lang_models = {}

    # We do this separately to give the caller a chance to set up logging
    def initialize(self):

        logging.info("Copying language schema files ...")
        self._copy_missing_language_files()

        # TODO: allow multiple language specifications
        logging.info("Loading language model ...")
        self._load_language_files()

    # https://stackoverflow.com/questions/1611799/preserve-case-in-configparser
    def optionxform(self, optionstr):
        return optionstr

    def getpath(self, section, option):
        return self._resolve_path(P(self.get(section, option)))

    def _resolve_path(self, path: P):
        """Paths in the config file can be absolute or relative. Absolute paths are left untouched
        relative paths are resolved relative to the configuration file location"""
        path = path.expanduser()
        if path.is_absolute():
            return path
        else:
            return P(self.cfg_path, path)

    def _copy_missing_language_files(self):
        for fn in ["schema-v1.0.json", "schema-v1.1.json", "schema-v1.2.0-dev1.json"]:
            src_file = P(default_config_data_dir, fn)
            dst_file = P(self.cfg_path, fn)
            if not dst_file.exists():
                shutil.copy(src_file, dst_file)

    def _load_language_files(self):
        for fname in self.cfg_path.glob("schema-*.json"):
            version = fname.name[7:-5]
            self.lang_models[version] = parse_schema(fname)
            logger.info(f"Loaded language schema {version}")



#  Copyright (c) 2019 Seven Bridges. See LICENSE

import time
import pathlib

from .yaml import parse_yaml
from .intelligence import Intelligence
from .intelligencecontext import IntelligenceContext
from ..cwl.specification import latest_published_cwl_version, process_types
from ..cwl.typeinference import infer_type
from .symbols import extract_symbols, extract_step_symbols
from .workflowgraph import cwl_graph
from ..langserver.lspobjects import Position

import logging
logger = logging.getLogger(__name__)


class Document:

    def __init__(self,
                 doc_uri: str,
                 scratch_path: pathlib.Path,  # Needed for ExecutionContext's example input file
                 text: str,
                 version: int,
                 type_dicts: dict):
        self.doc_uri = doc_uri
        self.config = scratch_path
        self.text = text
        self.version = version
        self.type_dicts = type_dicts

        self.problems = None
        self.code_intelligence = None
        self.symbols = None
        self.wf_graph = None

        self.update(text)

    def update(self, new_text):
        self.text = new_text
        self.code_intelligence = Intelligence()
        self.symbols = []

        t0 = time.time()
        cwl, self.problems = parse_yaml(self.text)
        t1 = time.time()
        logger.debug(f"Took {t1 - t0:1.3}s to load {self.doc_uri}")

        if not isinstance(cwl, dict):
            return

        t2 = time.time()
        self.code_intelligence.load_namespaces(cwl)
        self.code_intelligence.prepare_execution_context(self.doc_uri, cwl, self.config)

        self.parse(cwl)
        t3 = time.time()
        logger.debug(f"Took {t3 - t2:1.3}s to parse {self.doc_uri}")

        self.symbology(cwl)

    def definition(self, loc: Position):
        de = self.code_intelligence.get_doc_element(loc)
        if de is not None:
            return de.definition()

    def completion(self, loc: Position):
        de = self.code_intelligence.get_doc_element(loc)
        if de is not None:
            return de.completion()

    def hover(self, loc: Position):
        de = self.code_intelligence.get_doc_element(loc)
        if de is not None:
            return de.hover()

    def parse(self, cwl):
        cwl_v = cwl.get("cwlVersion")
        if cwl_v not in self.type_dicts:
            logger.error(f"No language model for cwl version {str(cwl_v)}. "
                         f"Using {latest_published_cwl_version}")
            cwl_v = latest_published_cwl_version

        lm = self.type_dicts.get(cwl_v)
        inferred_type = infer_type(
            node=cwl,
            allowed_types=[lm.get(t) for t in process_types])
        inferred_type.parse(
            doc_uri=self.doc_uri,
            node=cwl,
            intel_context=IntelligenceContext(path=[]),
            code_intel=self.code_intelligence,
            problems=self.problems)

    def symbology(self, cwl):
        line_count = self.text.count("\n")

        symbols = {}
        _typ = cwl.get("class")
        if _typ in process_types:
            symbols = extract_symbols(cwl, line_count)

            if _typ == "Workflow":
                symbols = extract_step_symbols(cwl, symbols)

        self.symbols = list(symbols.values())
        self.wf_graph = cwl_graph(cwl)



"""
Expected requests

    "initialize": self.serve_initialize,
    "initialized": self.serve_ignore,
    "textDocument/didOpen": self.serve_doc_did_open,
    "textDocument/didChange": self.serve_doc_did_change,
    "textDocument/completion": self.serve_completion,
    "textDocument/hover": self.serve_hover,
    "textDocument/codeAction": self.serve_available_commands,
    "textDocument/implementation":
    "textDocument/definition": self.serve_definition,
    "textDocument/xdefinition": self.serve_x_definition,
    "textDocument/references": self.serve_references,
    "workspace/xreferences": self.serve_x_references,
    "textDocument/documentSymbol": self.serve_document_symbols,
    "workspace/symbol": self.serve_symbols,
    "workspace/xpackages": self.serve_x_packages,
    "workspace/xdependencies": self.serve_x_dependencies,
    "$/cancelRequest": self.serve_ignore,
    "shutdown": self.serve_ignore,
    "exit": self.serve_exit,

"""
#  Copyright (c) 2019 Seven Bridges. See LICENSE

from enum import IntEnum

from .lspobjects import to_dict
from .base import CWLLangServerBase, JSONRPC2Error, ServerError, LSPErrCode
from .fileoperation import FileOperation
from .definition import Definition
from .completion import Completion
from .documentsymbol import DocumentSymbol
from .hover import Hover

import logging

logger = logging.getLogger(__name__)
logger.propagate = True

logging.getLogger("benten.langserver.jsonrpc").propagate = False


class TextDocumentSyncKind(IntEnum):
    _None = 0
    Full = 1
    Incremental = 2


class LangServer(
        Hover,
        DocumentSymbol,
        Completion,
        Definition,
        FileOperation,
        CWLLangServerBase):

    def run(self):
        while self.running:
            try:
                request = self.conn.read_message()
                self.handle(request)
            except EOFError:
                break
            except Exception as e:
                logger.error("Unexpected error: %s", e, exc_info=True)

    # Request message:
    # {
    # 	"jsonrpc": "2.0",
    # 	"id": 1, # <----------  if this field is missing, it's a notification
    #                       	and should not receive a response. Otherwise,
    #                           the response should carry this id for accounting
    #                           purposes
    # 	"method": "...",
    # 	"params": {
    # 		...
    # 	}
    # }
    def handle(self, client_query):
        logger.info("Client query: {}".format(client_query.get("method")))

        is_a_request = "id" in client_query

        if self.premature_request(client_query, is_a_request):
            return

        if self.duplicate_initialization(client_query, is_a_request):
            return

        try:
            response = to_dict(self._dispatch(client_query))

            if is_a_request:
                self.conn.write_response(client_query["id"], response)

        except ServerError as e:
            logger.error(e.server_error_message)

            let_client_know_of_errors = False
            if let_client_know_of_errors:
                self.conn.write_error(
                    client_query["id"],
                    code=e.json_rpc_error.code,
                    message=str(e.json_rpc_error.message),
                    data=e.json_rpc_error.data)

    def premature_request(self, client_query, is_a_request):
        if not self.initialization_request_received and \
                client_query.get("method", None) not in ["initialize", "exit"]:
            logger.warning("Client sent a request/notification without initializing")
            if is_a_request:
                self.conn.write_error(
                    client_query["id"],
                    code=LSPErrCode.ServerNotInitialized,
                    message="",
                    data={})
            return True
        else:
            return False

    def duplicate_initialization(self, client_query, is_a_request):
        if self.initialization_request_received and client_query.get("method", None) == "initialize":
            logger.warning("Client sent duplicate initialization")
            if is_a_request:
                self.conn.write_error(
                    client_query["id"],
                    code=LSPErrCode.InvalidRequest,
                    message="Client sent duplicate initialization",
                    data={})
            return True
        else:
            return False

    def _dispatch(self, client_query):
        # textDocument/didOpen -> serve_textDocument_didOpen
        method_name = "serve_" + client_query.get("method", "noMethod").replace("/", "_")
        try:
            f = getattr(self, method_name)
        except AttributeError as e:
            f = self.serve_unknown
        return f(client_query)

    @staticmethod
    def serve_noMethod(client_query):
        msg = "Method not specified"
        raise ServerError(
            server_error_message=msg,
            json_rpc_error=JSONRPC2Error(
                code=LSPErrCode.MethodNotFound,
                message=msg))

    @staticmethod
    def serve_unknown(client_query):
        msg = "Unknown method: {}".format(client_query["method"])
        raise ServerError(
            server_error_message=msg,
            json_rpc_error=JSONRPC2Error(
                code=LSPErrCode.MethodNotFound,
                message=msg))

    def serve_shutdown(self, client_query):
        logging.shutdown()
        self.running = False

    # https://microsoft.github.io/language-server-protocol/specification#initialize
    def serve_initialize(self, client_query):
        self.initialization_request_received = True

        self.client_capabilities = client_query.get("capabilities", {})
        logger.debug("InitOpts: {}".format(client_query))

        return {
            "capabilities": {
                "textDocumentSync": TextDocumentSyncKind.Full,
                #  Avoid complexity of incremental updates for now
                "completionProvider": {
                    "resolveProvider": True,
                    "triggerCharacters": [".", "/"]
                },
                "hoverProvider": True,
                "definitionProvider": True,
                "referencesProvider": True,
                "documentSymbolProvider": True,
                "workspaceSymbolProvider": True,
                "streaming": True,
                "codeActionProvider": {
                    "codeActionKinds": ["source"]
                },
                "workspace": {
                    "workspaceFolders": {
                        "supported": True,
                        "changeNotifications": True
                    }
                },
                # https://github.com/sourcegraph/language-server-protocol/blob/master/extension-files.md#files-extensions-to-lsp
                # This is not in the spec yet
                "xfilesProvider": True
            }
        }

    @staticmethod
    def serve_initialized(client_query):
        return {}




"""Iterates over the CWL and extracts top level and step symbols"""

#  Copyright (c) 2019 Seven Bridges. See LICENSE

import textwrap

from ..langserver.lspobjects import SymbolKind, DocumentSymbol, Position, Range


def truncate(text):
    if isinstance(text, str):
        if len(text):
            return textwrap.shorten(text, 20, placeholder="...")
    return "-"


# SymbolKind has a nice representation in the VS Code UI
Symbols = {
    "class": lambda k, v: {
        "kind": SymbolKind.File,
        "name": k,
        "detail": v
    },
    "cwlVersion": lambda k, v: {
        "kind": SymbolKind.Constant,
        "name": k,
        "detail": v
    },
    "id": lambda k, v: {
        "kind": SymbolKind.String,
        "name": truncate(v)
    },
    "label": lambda k, v: {
        "kind": SymbolKind.String,
        "name": truncate(v)
    },
    "doc": lambda k, v: {
        "kind": SymbolKind.String,
        "name": truncate(v)
    },
    "inputs": lambda k, v: {
        "kind": SymbolKind.Interface,
        "name": k
    },
    "outputs": lambda k, v: {
        "kind": SymbolKind.Interface,
        "name": k
    },
    "baseCommand": lambda k, v: {
        "kind": SymbolKind.Operator,
        "name": k
    },
    "expression": lambda k, v: {
        "kind": SymbolKind.Function,
        "name": "{}"
    },
    "requirements": lambda k, v: {
        "kind": SymbolKind.Array,
        "name": k
    },
    "hints": lambda k, v: {
        "kind": SymbolKind.Array,
        "name": k
    },
    "steps": lambda k, v: {
        "kind": SymbolKind.Class,
        "name": k
    }
}


def SymbolDefault(k, v):
    return {
        "kind": SymbolKind.Field,
        "name": k
    }


def extract_symbols(cwl, line_count):
    starts = [cwl.lc.key(k)[0] for k in cwl.keys()]
    starts += [line_count - 1]
    return {
        k: DocumentSymbol(
                _range=Range(
                    start=Position(starts[n], 0),
                    end=Position(starts[n + 1] - 1, 0)
                ),
                selection_range=Range(
                    start=Position(starts[n], 0),
                    end=Position(starts[n + 1] - 1, 0)
                ),
                **Symbols.get(k, SymbolDefault)(k, cwl[k])
            )
        for n, k in enumerate(cwl.keys())
    }


def extract_step_symbols(cwl, symbols):
    _steps = cwl.get("steps")
    if not isinstance(_steps, (list, dict)):
        return symbols

    symb_steps = symbols.get("steps")
    last_line = symb_steps.range.end.line

    if isinstance(_steps, list):
        lines = [_steps.lc.item(n)[0] for n in range(len(_steps))]
        names = []
        for n, v in enumerate(_steps):
            mid = f"Step {n} is missing id!"
            if isinstance(v, dict):
                name = v.get("id", mid)
            else:
                name = mid
            names += [name]
    else:
        lines = [_steps.lc.key(k)[0] for k in _steps.keys()]
        names = [k for k in _steps.keys()]

    lines += [last_line]

    symb_steps.children = [
        DocumentSymbol(
            name=names[n],
            kind=SymbolKind.Field,
            _range=Range(
                start=Position(lines[n], 0),
                end=Position(lines[n+1] - 1, 0)
            ),
            selection_range=Range(
                start=Position(lines[n], 0),
                end=Position(lines[n+1] - 1, 0)
            )
        )
        for n in range(len(names))
    ]

    return symbols



#  Copyright (c) 2019 Seven Bridges. See LICENSE


from .basetype import CWLBaseType, MapSubjectPredicate
from ..code.workflow import Workflow
from ..code.intelligence import Intelligence
from ..langserver.lspobjects import Range, Diagnostic, DiagnosticSeverity

import logging
logger = logging.getLogger(__name__)


class CWLUnknownType(CWLBaseType):

    def __init__(self, name, expected):
        super().__init__(name)
        self.expected = expected

    def parse(self,
              doc_uri: str,
              node,
              intel_context: Workflow,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):

        problems += [
            Diagnostic(
                _range=value_range,
                message=f"Got type {self.name}. Expecting one of {self.expected}",
                severity=DiagnosticSeverity.Error)
        ]



"""Provides classes that can be used to infer document location
from cursor location and provide auto-completions.

Language Model objects parse the document to produce document
model objects.

For details see ../../docs/document-model.md
"""

#  Copyright (c) 2019 Seven Bridges. See LICENSE

from typing import List
import pathlib

from ..langserver.lspobjects import (Position, Range, CompletionItem)
from .executioncontext import ExecutionContext

import logging
logger = logging.getLogger(__name__)


class LookupNode:

    def __init__(self, loc: Range):
        self.loc = loc
        self.intelligence_node = None


class IntelligenceNode:

    def __init__(self, completions: List[str]=None):
        self._completions = completions or []

    def completion(self):
        return [CompletionItem(label=c) for c in self._completions]

    def hover(self):
        pass

    def definition(self):
        pass


# # Children of maps need to track their ancestors for completions
# class IntelligenceContext:
#     pass


class Intelligence:

    def __init__(self):
        self.lookup_table: List[LookupNode] = []
        self.type_defs = {}
        self.namespaces = {}
        self.execution_context: ExecutionContext = None

    def add_lookup_node(self, node: LookupNode):
        self.lookup_table.append(node)

    def load_namespaces(self, cwl: dict):
        if "$namespaces" in cwl:
            self.namespaces = cwl["$namespaces"]

    def prepare_execution_context(self, doc_uri: str, cwl: dict, scratch_path: pathlib.Path):
        self.execution_context = ExecutionContext(
            doc_uri=doc_uri,
            scratch_path=scratch_path,
            cwl=cwl,
            user_types=self.type_defs)

    def prepare_expression_lib(self, expression_lib: list):
        self.execution_context.set_expression_lib(expression_lib)

    def get_doc_element(self, loc: Position):
        # O(n) algorithm, but should do fine for our file sizes
        # For now doing exact matches on lines, which is sufficient
        for n in self.lookup_table:
            if n.loc.start.line <= loc.line <= n.loc.end.line:
                if loc.line > n.loc.start.line or loc.character >= n.loc.start.character:
                    if loc.line < n.loc.end.line or loc.character <= n.loc.end.character:
                        return n.intelligence_node

        return None



#  Copyright (c) 2019 Seven Bridges. See LICENSE

from typing import Dict
from enum import IntEnum

from ..code.document import Document

import logging

logger = logging.getLogger(__name__)
logger.propagate = True


# https://stackoverflow.com/a/24482131
class LSPErrCode(IntEnum):
    # Defined by JSON RPC
    ParseError = -32700
    InvalidRequest = -32600
    MethodNotFound = -32601
    InvalidParams = -32602
    InternalError = -32603
    serverErrorStart = -32099
    serverErrorEnd = -32000
    ServerNotInitialized = -32002
    UnknownErrorCode = -32001

    # Defined by the protocol.
    RequestCancelled = -32800
    ContentModified = -32801


class JSONRPC2Error(Exception):
    def __init__(self, code, message, data=None):
        self.code = code
        self.message = message
        self.data = data


class ServerError(Exception):
    def __init__(self, server_error_message, json_rpc_error):
        self.server_error_message: str = server_error_message
        self.json_rpc_error: JSONRPC2Error = json_rpc_error


class CWLLangServerBase:
    def __init__(self, conn, config):
        self.conn = conn
        self.running = True
        self.root_path = None
        self.fs = None
        self.all_symbols = None
        self.workspace = None
        self.streaming = True

        self.open_documents: Dict[str, Document] = {}
        self.initialization_request_received = False

        self.client_capabilities = {}

        self.config = config



#  Copyright (c) 2019 Seven Bridges. See LICENSE

from typing import List

from .basetype import CWLBaseType, MapSubjectPredicate, TypeCheck, Match
from .unknowntype import CWLUnknownType
from .anytype import CWLAnyType
from .namespacedtype import CWLNameSpacedType


def infer_type(node, allowed_types,
               key: str = None, map_sp: MapSubjectPredicate = None) -> CWLBaseType:
    type_check_results = check_types(node, allowed_types, key, map_sp)
    for tcr in type_check_results:
        if tcr.match == Match.Yes:
            res = tcr.cwl_type
            break
    else:
        for tcr in type_check_results:
            if tcr.match == Match.Maybe:
                res = tcr.cwl_type
                break
        else:
            if len(type_check_results) == 1:
                res = type_check_results[0].cwl_type
            else:
                res = CWLUnknownType(name="(unknown)",
                                     expected=[tr.cwl_type.name for tr in type_check_results])
    return res


def check_types(node, allowed_types, key, map_sp) -> List[TypeCheck]:

    type_check_results = []

    explicit_type = get_explicit_type_str(node, key, map_sp)

    if explicit_type is not None:
        if ":" in explicit_type:
            return [TypeCheck(CWLNameSpacedType(explicit_type))]

        for _type in allowed_types:
            if isinstance(_type, CWLAnyType):
                req_type = _type.if_you_can_be_anything_be_this_kind(explicit_type)
                if req_type is not None:
                    return [TypeCheck(req_type)]
                else:
                    return [TypeCheck(
                        CWLUnknownType(name=explicit_type,
                                       expected=_type.all_possible_type_names()),
                        match=Match.No)]

            if explicit_type == _type.name:
                return [TypeCheck(_type)]

        return [
            TypeCheck(CWLUnknownType(
                name=explicit_type,
                expected=[t.name for t in allowed_types]),
                match=Match.No)]

    for _type in allowed_types:
        if _type.name == 'null':
            if node is None:
                return [TypeCheck(CWLBaseType(name=_type))]
            else:
                type_check_results += [TypeCheck(CWLBaseType(name=_type), match=Match.No)]
                continue

        # string greedily matches Expression, so we have to take care of this ...
        if _type.name == "string":
            if node is None:
                return [TypeCheck(_type)]

            elif isinstance(node, str):
                type_check_results += [TypeCheck(_type, match=Match.Maybe)]

            else:
                type_check_results += [TypeCheck(_type, match=Match.No)]

            continue

        # For now, don't closely validate these base types
        if _type.name in ['boolean', 'int', 'long']:
            if node is None or isinstance(node, (str, bool, int)):
                return [TypeCheck(_type)]
            else:
                type_check_results += [TypeCheck(_type, match=Match.No)]
                continue

        check_result = _type.check(node, key, map_sp)
        if check_result.match == Match.Yes:
            return [check_result]

        type_check_results += [check_result]

    return type_check_results


def get_explicit_type_str(node, key: str, map_sp: MapSubjectPredicate):
    if map_sp is not None:
        if map_sp.subject == "class":
            return key
        else:
            return None
    elif isinstance(node, dict):
        return node.get("class")
    else:
        return None



#  Copyright (c) 2019 Seven Bridges. See LICENSE

import sys
import socketserver
import pathlib

from .configuration import Configuration

from benten.version import __version__
from benten.langserver.jsonrpc import JSONRPC2Connection, ReadWriter, TCPReadWriter
from benten.langserver.server import LangServer

from logging.handlers import RotatingFileHandler
import logging.config
import logging
logger = logging.getLogger()


# https://stackoverflow.com/a/47075664
# https://docs.python.org/3.6/library/socketserver.html#socketserver.ForkingMixIn
class ForkingTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    pass


class LangserverTCPTransport(socketserver.StreamRequestHandler):

    config = None

    def handle(self):
        conn = JSONRPC2Connection(TCPReadWriter(self.rfile, self.wfile))
        s = LangServer(conn=conn, config=self.config)
        s.run()


def main():
    import argparse

    config = Configuration()

    log_fn = pathlib.Path(config.log_path, "benten-ls.log")
    roll_over = log_fn.exists()

    handler = RotatingFileHandler(log_fn, backupCount=5)
    formatter = logging.Formatter(
        fmt='[%(levelname)-7s] %(asctime)s (%(name)s) %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    handler.setFormatter(formatter)
    if roll_over:
        handler.doRollover()

    # logging.basicConfig(filename=log_fn, filemode="w", level=logging.INFO)
    # logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser(description="")
    parser.add_argument(
        "--mode", default="stdio", help="communication (stdio|tcp)")
    parser.add_argument(
        "--addr", default=4389, help="server listen (tcp)", type=int)
    parser.add_argument("--debug", action="store_true")

    args = parser.parse_args()

    logging.basicConfig(level=(logging.DEBUG if args.debug else logging.WARNING))
    logger.addHandler(handler)

    logger.info(f"Benten {__version__}: CWL Language Server from Rabix (Seven Bridges)")

    config.initialize()

    if args.mode == "stdio":
        logger.info("Reading on stdin, writing on stdout")
        s = LangServer(
            conn=JSONRPC2Connection(ReadWriter(sys.stdin.buffer, sys.stdout.buffer)),
            config=config)
        s.run()
    elif args.mode == "tcp":
        host, addr = "0.0.0.0", args.addr
        logger.info("Accepting TCP connections on %s:%s", host, addr)
        ForkingTCPServer.allow_reuse_address = True
        ForkingTCPServer.daemon_threads = True
        LangserverTCPTransport.config = config
        s = ForkingTCPServer((host, addr), LangserverTCPTransport)
        try:
            s.serve_forever()
        finally:
            s.shutdown()


if __name__ == "__main__":
    main()



#  Copyright (c) 2019 Seven Bridges. See LICENSE

from ruamel.yaml import YAML
from ruamel.yaml.parser import ParserError
from ruamel.yaml.scanner import ScannerError

from ..cwl.lib import resolve_file_path


fast_load = YAML(typ='safe')


def extract_schemadef(doc_uri: str, cwl: dict):
    _req = cwl.get("requirements")
    _types = []
    types_dict = {}
    if isinstance(_req, dict):
        if "SchemaDefRequirement" in _req:
            _types = _req.get("SchemaDefRequirement", {}).get("types")
    elif isinstance(_req, list):
        for _this_req in _req:
            if isinstance(_this_req, dict):
                if _this_req.get("class") == "SchemaDefRequirement":
                    _types = _this_req.get("types")

    if isinstance(_types, list):
        for _type in _types:
            if isinstance(_type, dict):
                name = None
                if list(_type.keys()) == ["$import"]:
                    path = _type.get("$import")
                    _type = load_typedefs_from_file(doc_uri, path)
                    if isinstance(_type, dict):
                        if "name" in _type:
                            name = path + "#" + _type.pop("name")
                else:
                    name = _type.pop("name", None)

                if name is not None:
                    types_dict[name] = _type

    return types_dict


def load_typedefs_from_file(doc_uri, path):
    linked_file = resolve_file_path(doc_uri, path)
    type_def = {}
    if linked_file.exists() and linked_file.is_file():
        try:
            type_def = fast_load.load(linked_file.open("r").read())
        except (ParserError, ScannerError) as e:
            type_def = {}
            # todo: flag errors in imported typedefs
    else:
        pass
        # The missing file error should already be flagged by the main parse

    return type_def




"""Load the raw YAML"""

#  Copyright (c) 2019 Seven Bridges. See LICENSE

from typing import Tuple, List

from ruamel.yaml import YAML
from ruamel.yaml.parser import ParserError
from ruamel.yaml.scanner import ScannerError
from ruamel.yaml.composer import ComposerError
from ruamel.yaml.compat import StringIO

from ..langserver.lspobjects import Diagnostic, DiagnosticSeverity, Range, Position


import logging
logger = logging.getLogger(__name__)

_yaml_loader = YAML(typ="rt")
# TODO: allow checking for duplicate keys, perhaps with self healing
_yaml_loader.allow_duplicate_keys = True

fast_load = YAML(typ='safe')
fast_load.indent(mapping=2, sequence=4, offset=2)
fast_load.default_flow_style = False


def fast_yaml_load(txt):
    try:
        return fast_load.load(txt)
    except (ParserError, ScannerError) as e:
        pass


def yaml_to_string(v: dict):
    s = StringIO()
    fast_load.dump(v, s)
    return s.getvalue()


def parse_yaml(text, retries=3) -> Tuple[dict, List[Diagnostic]]:
    problems = []
    try:
        cwl = _yaml_loader.load(text)
    except (ParserError, ScannerError, ComposerError) as e:

        if retries:

            # TODO: refactor - cleanup healing code
            if e.problem == "could not find expected ':'":
                return parse_yaml(heal_incomplete_key(text, e), retries - 1)

            if e.problem == "mapping values are not allowed here":
                return parse_yaml(heal_incomplete_key_typeB(text, e), retries - 1)

        cwl = None
        problems = [
            Diagnostic(
                _range=Range(start=Position(e.problem_mark.line, e.problem_mark.column),
                             end=Position(e.problem_mark.line, e.problem_mark.column)),
                message=str(e),
                severity=DiagnosticSeverity.Error,
                code="YAML err",
                source="Benten")]

    return cwl, problems


def heal_incomplete_key(original_text, e):
    logger.debug("Attempting to heal incomplete key")
    lines = original_text.splitlines(keepends=False)
    # TODO: This only works for block style, but it does the job
    lines[e.context_mark.line] = lines[e.context_mark.line] + ":"
    return "\n".join(lines)


def heal_incomplete_key_typeB(original_text, e):
    logger.debug("Attempting to heal incomplete key")
    lines = original_text.splitlines(keepends=False)
    # Walk back up to first non-empty line and slap a ":" on the end
    ln = e.problem_mark.line - 1
    while ln > 0 and len(lines[ln].strip()) == 0:
        ln -= 1

    if len(lines[ln]):
        lines[ln] = lines[ln] + ":"

    return "\n".join(lines)



#  Copyright (c) 2019 Seven Bridges. See LICENSE

import shlex
import pathlib

from .basetype import CWLBaseType, IntelligenceContext, Intelligence, MapSubjectPredicate
from ..langserver.lspobjects import Range, Location, CompletionItem, Hover
from .lib import validate_and_load_linked_file
from ..code.intelligence import LookupNode

import logging
logger = logging.getLogger(__name__)


class CWLLinkedFile(CWLBaseType):

    def __init__(self, prefix, extension=None):
        super().__init__("Linked file")
        self.prefix = prefix
        self.full_path = None
        self._contents = None
        self.node_dict = None
        self.extension = extension

    def parse(self,
              doc_uri: str,
              node,
              intel_context: IntelligenceContext,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):

        self.full_path, self._contents, self.node_dict = \
            validate_and_load_linked_file(doc_uri, self.prefix, value_range, problems)
        ln = LookupNode(loc=value_range)
        ln.intelligence_node = self
        code_intel.add_lookup_node(ln)

    def hover(self):
        return Hover(self._contents)

    def definition(self):
        if isinstance(self.full_path, pathlib.Path):
            return Location(self.full_path.as_uri())

    def completion(self):
        return [
            CompletionItem(label=file_path)
            for file_path in self._file_picker()
        ]

    def _file_picker(self):

        # This is an http path, not a file system path
        if not isinstance(self.full_path, pathlib.Path):
            return []

        # We use .split( ... ) so we can handle the case for run: .    # my/commented/path
        # and other such shenanigans
        _prefix = shlex.split(self.prefix, comments=True)[0]

        path = self.full_path
        if not path.is_dir():
            path = path.parent

        if not path.exists():
            logger.error(f"No path called: {path}")
            return []

        # This is a workaround to the issue of having a dangling "." in front of the path
        pre = "/" if _prefix in [".", ".."] else ""

        return (pre + str(p.relative_to(path))
                for p in path.iterdir()
                if p.is_dir() or (self.extension is None or p.suffix == self.extension))



#  Copyright (c) 2019 Seven Bridges. See LICENSE

from .linkedfiletype import CWLLinkedFile
from .basetype import IntelligenceContext, Intelligence, MapSubjectPredicate
from ..langserver.lspobjects import Range

import logging
logger = logging.getLogger(__name__)


class CWLLinkedSchemaDef(CWLLinkedFile):

    def parse(self,
              doc_uri: str,
              node,
              intel_context: IntelligenceContext,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):
        super().parse(doc_uri, node, intel_context, code_intel, problems, node_key, map_sp, key_range, value_range, requirements)
        if isinstance(self.node_dict, dict):
            if "name" in self.node_dict:
                name = self.prefix + "#" + self.node_dict.pop("name")
                code_intel.type_defs[name] = self.node_dict



#  Copyright (c) 2019 Seven Bridges. See LICENSE

import re
from enum import IntEnum

import dukpy

from .basetype import (CWLBaseType, MapSubjectPredicate, TypeCheck, Match,
                       Intelligence, IntelligenceContext)
from ..langserver.lspobjects import Range, Hover, Location
from ..code.intelligence import LookupNode

import logging
logger = logging.getLogger(__name__)


parameter_ref = re.compile(r"\$\(((.(?<!\$({|\()))*)\)", flags=re.DOTALL | re.M)
expression_ref = re.compile(r"\${((.(?<!\$({|\()))*)}", flags=re.DOTALL | re.M)

inputs_scan = re.compile(r"inputs\.([\w]*)", flags=re.DOTALL | re.M)


class CWLExpressionType(CWLBaseType):

    def check(self, node, node_key: str=None, map_sp: MapSubjectPredicate=None) -> TypeCheck:
        if isinstance(node, str):
            if "$(" in node or "${" in node:
                return TypeCheck(cwl_type=CWLExpression(node))

        return TypeCheck(cwl_type=self, match=Match.No)


class ExpressionType(IntEnum):
    PlainString = 0
    ParameterReference = 1
    JSExpression = 2


class CWLExpression(CWLBaseType):

    def __init__(self, text: str):
        super().__init__("Expression")
        self.text = text
        self.intel_context = None
        self.execution_context = None
        self.range = None

    def guess_inputs(self):
        return [inp.groups()[0] for inp in inputs_scan.finditer(self.text)]

    def parse(self,
              doc_uri: str,
              node,
              intel_context: IntelligenceContext,
              code_intel: Intelligence,
              problems: list,
              node_key: str = None,
              map_sp: MapSubjectPredicate = None,
              key_range: Range = None,
              value_range: Range = None,
              requirements=None):

        self.intel_context = intel_context
        self.execution_context = code_intel.execution_context
        self.range = value_range  # For the highlighting of the expression

        ln = LookupNode(loc=value_range)
        ln.intelligence_node = self
        code_intel.add_lookup_node(ln)

    def hover(self):

        def _self_is_io(_path):
            if "in" in _path:
                return False

            for _component in ["secondaryFiles", "position", "valueFrom"]:
                if _component in _path:
                    return True

            return False

        def _self_is_outputEval(_path):
            return "in" not in _path and "outputEval" in _path

        def _self_is_in_step(_path):
            if "in" in _path and "valueFrom" in _path:
                return True
            else:
                return False

        job_inputs = self.execution_context.sample_data["inputs"]
        job_outputs = self.execution_context.sample_data["outputs"]
        cwl_self = None

        # Deal with filling out self
        try:
            if _self_is_io(self.intel_context.path):
                if "inputs" in self.intel_context.path:
                    cwl_self = job_inputs.get(self.intel_context.path[1])
                elif "outputs" in self.intel_context.path:
                    cwl_self = job_outputs[self.intel_context.path[1]]

            elif _self_is_outputEval(self.intel_context.path):
                # todo: Need to check for `glob`
                cwl_self = self.execution_context.get_sample_globbed_files(self.intel_context.path[1])

            elif _self_is_in_step(self.intel_context.path):
                job_inputs, cwl_self = \
                    self.execution_context.get_workflow_step_inputs(self.intel_context.path)

        except (ValueError, IndexError) as e:
            pass

        if job_inputs:
            res = "".join(
                evaluate_expression(
                    expression=fragment["exp"],
                    exp_type=fragment["type"],
                    expression_lib=self.execution_context.expression_lib,
                    runtime=self.execution_context.runtime(self.intel_context.path),
                    inputs=job_inputs,
                    cwl_self=cwl_self)
                for fragment in self._split_fragments())
        else:
            res = "Job inputs have not been filled out"

        logger.debug(f"Guessing expression inputs are: {self.guess_inputs()}")
        return Hover(res, self.range)

    def definition(self):
        # Hijacking this to show the sample inputs file
        return Location(self.execution_context.get_sample_data_file_path().as_uri())

    def _split_fragments(self) -> list:
        refs = parameter_ref.finditer(self.text)
        exps = expression_ref.finditer(self.text)
        r, e = next(refs, None), next(exps, None)

        cursor = 0
        fragments = []
        while r or e:
            if r is not None and e is not None:
                if r.start() < e.start():
                    _frag = self._add_ref(r)
                    r = next(refs, None)
                else:
                    _frag = self._add_exp(e)
                    e = next(exps, None)
            else:
                if r is not None:
                    _frag = self._add_ref(r)
                    r = next(refs, None)
                elif e is not None:
                    _frag = self._add_exp(e)
                    e = next(exps, None)

            plain_string_frag = self._add_plain_string(self.text, (cursor, _frag["span"][0]))
            cursor = _frag["span"][1]
            fragments += [plain_string_frag, _frag]

        plain_string_frag = self._add_plain_string(self.text, (cursor, len(self.text)))
        fragments += [plain_string_frag]
        return fragments

    @staticmethod
    def _add_plain_string(text, span):
        return {
            "exp": text[span[0]:span[1]],
            "type": ExpressionType.PlainString,
            "span": span
        }

    @staticmethod
    def _add_ref(r):
        return {
            "exp": r.groups()[0],
            "type": ExpressionType.ParameterReference,
            "span": r.span()
        }

    @staticmethod
    def _add_exp(e):
        return {
            "exp": e.groups()[0],
            "type": ExpressionType.JSExpression,
            "span": e.span()
        }


def parameter_reference_template(expression):
    return f"""
var runtime = dukpy['runtime'];
var inputs = dukpy['inputs'];
var self = dukpy['cwl_self'];
{expression}"""


def js_template(expression):
    return f"""
var runtime = dukpy['runtime'];
var inputs = dukpy['inputs'];
var self = dukpy['cwl_self'];
function benten_eval_func() {{
    {expression} 
}};
benten_eval_func()"""


def evaluate_expression(
        expression: str, exp_type: ExpressionType,
        expression_lib: list, runtime: dict, inputs: dict, cwl_self: dict):
    if exp_type == ExpressionType.PlainString:
        return expression

    if inputs:
        if exp_type == ExpressionType.ParameterReference:
            full_expression = parameter_reference_template(expression)
        else:
            full_expression = js_template(expression)

        try:
            res = dukpy.evaljs(expression_lib + [full_expression],
                               runtime=runtime,
                               inputs=inputs,
                               cwl_self=cwl_self)

            if res is None and exp_type == ExpressionType.JSExpression:
                res = "Got a 'null' result. Do you have a `return` for your JS expression?"
            else:
                res = str(res)

        except dukpy.JSRuntimeError as e:
            res = str(e).splitlines()[0]
            logger.error(res)
    else:
        res = "Job inputs have not been filled out"

    return res



#  Copyright (c) 2019 Seven Bridges. See LICENSE

import pathlib
from datetime import datetime
from setuptools import setup, find_packages

current_path = pathlib.Path(__file__).parent

name = 'benten'
ver_path = pathlib.Path(current_path, "benten", "version.py")
_ver = {}
exec(ver_path.open("r").read(), _ver)
version = _ver["__version__"]
now = datetime.utcnow()
desc_path = pathlib.Path(current_path, "Readme.md")
long_description = desc_path.open("r").read()

setup(
    name=name,
    version=version,
    packages=find_packages(),
    platforms=['POSIX', 'MacOS', 'Windows'],
    python_requires='>=3.7.0',
    install_requires=[
        "ruamel.yaml >= 0.15.77",
        "dukpy >= 0.2.2"
    ],
    entry_points={
        'console_scripts': [
            'benten-ls = benten.__main__:main'
        ],
    },

    author='Seven Bridges Genomics Inc.',
    maintainer='Seven Bridges Genomics Inc.',
    maintainer_email='kaushik.ghose@sbgenomics.com',
    author_email='kaushik.ghose@sbgenomics.com',
    description='CWL language server developed by Seven Bridges',
    url='https://github.com/rabix/benten',
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    license='Copyright (c) {} Seven Bridges'.format(now.year),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only'
    ],
    keywords='seven bridges cwl common workflow language'
)
